# from fontTools.misc.textTools import deHexStr, hexStr
# from io import BytesIO
from unittest import TestCase

from pytest import raises

from vfbLib.compilers.base import StreamWriter


class StreamWriterTest(TestCase):
    # uint8

    def test_write_uint8_0(self):
        s = StreamWriter()
        s.write_uint8(0)
        assert s.stream.getvalue() == b"\00"

    def test_write_uint8_128(self):
        s = StreamWriter()
        s.write_uint8(128)
        assert s.stream.getvalue() == b"\x80"

    def test_write_uint8_max(self):
        s = StreamWriter()
        s.write_uint8(255)
        assert s.stream.getvalue() == b"\xff"

    def test_write_uint8_negative(self):
        s = StreamWriter()
        with raises(OverflowError):
            s.write_uint8(-1)

    def test_write_uint8_large(self):
        s = StreamWriter()
        with raises(OverflowError):
            s.write_uint8(256)

    # uint16

    def test_write_uint16_0(self):
        s = StreamWriter()
        s.write_uint16(0)
        assert s.stream.getvalue() == b"\x00\x00"

    def test_write_uint16_128(self):
        s = StreamWriter()
        s.write_uint16(128)
        assert s.stream.getvalue() == b"\x80\x00"  # LE

    def test_write_uint16_max(self):
        s = StreamWriter()
        s.write_uint16(65535)
        assert s.stream.getvalue() == b"\xff\xff"

    def test_write_uint16_negative(self):
        s = StreamWriter()
        with raises(OverflowError):
            s.write_uint16(-1)

    def test_write_uint16_large(self):
        s = StreamWriter()
        with raises(OverflowError):
            s.write_uint16(65536)

    # int16 (signed)

    def test_write_int16_0(self):
        s = StreamWriter()
        s.write_int16(0)
        assert s.stream.getvalue() == b"\x00\x00"

    def test_write_int16_128(self):
        s = StreamWriter()
        s.write_int16(128)
        assert s.stream.getvalue() == b"\x80\x00"  # LE

    def test_write_int16_max(self):
        s = StreamWriter()
        s.write_int16(32767)
        assert s.stream.getvalue() == b"\xff\x7f"

    def test_write_int16_negative(self):
        s = StreamWriter()
        s.write_int16(-1)
        assert s.stream.getvalue() == b"\xff\xff"

    def test_write_int16_negative_max(self):
        s = StreamWriter()
        s.write_int16(-32767)
        assert s.stream.getvalue() == b"\x01\x80"

    def test_write_int16_large(self):
        s = StreamWriter()
        with raises(OverflowError):
            s.write_int16(32768)

    def test_write_int16_large_negative(self):
        s = StreamWriter()
        with raises(OverflowError):
            s.write_int16(-32769)

    # int32 (signed)

    def test_write_int32_0(self):
        s = StreamWriter()
        s.write_int32(0)
        assert s.stream.getvalue() == b"\x00\x00\x00\x00"

    def test_write_int32_128(self):
        s = StreamWriter()
        s.write_int32(128)
        assert s.stream.getvalue() == b"\x80\x00\x00\x00"  # LE

    def test_write_int32_max(self):
        s = StreamWriter()
        s.write_int32(2147483647)
        assert s.stream.getvalue() == b"\xff\xff\xff\x7f"

    def test_write_int32_negative(self):
        s = StreamWriter()
        s.write_int32(-1)
        assert s.stream.getvalue() == b"\xff\xff\xff\xff"

    def test_write_int32_negative_max(self):
        s = StreamWriter()
        s.write_int32(-2147483648)
        assert s.stream.getvalue() == b"\x00\x00\x00\x80"

    def test_write_int32_large(self):
        s = StreamWriter()
        with raises(OverflowError):
            s.write_int32(2147483648)

    def test_write_int32_large_negative(self):
        s = StreamWriter()
        with raises(OverflowError):
            s.write_int32(-2147483649)



from io import BytesIO
from pprint import pprint
from sys import argv

from fontTools.misc.textTools import deHexStr, hexStr

from vfbLib.parsers import EncodedValueParser, MetricsParser
from vfbLib.parsers.glyph import GlyphParser
from vfbLib.parsers.mm import PrimaryInstancesParser


def decode_data(data):
    result = EncodedValueParser.parse(BytesIO(data), len(data))
    print(result)


def decode_glyph():
    data = deHexStr(
        "01090701019041452e7363088df77aa200f7b38bf7ef8b018bf71a8bd201fb538bfb1e8b0140fb1a6a4401528bfb548b01f7c0f89cf790f89c01f7c08bf86b8b018b618bfb2501fb738bfb478b018bfb508b6101f7608bf7378b018b618bfb1f01fb608bfb378b018bfb618b5c01f7738bf7478b018b608bfb2700fba9f86bfbfcf80b01898b878b048177837504735e7657017666817601fb05fb58643601f73e8bdd8b02f9128bf9888b038b8ba0018c028b018b018d028cff8c018e028b018b028cff8e018c028c018b028b018dffa0018e028b018b028c048cf80b8bf80b8b8c8b8b8b8b0af703a2078f8b038f99898a0d8b8f998b038f8e8b8a038b908a8a0d9c908b8a0d8d8e9c8a0d928f998b038b988b8a0398948a8a0d96998f8b0da09c8e8a038ba18a8a019092028b8c028e8c0e8c8b908b0490928b8a0e978b908b0497948b8a048b988b8a0e9c8b908a048ca18b8a8b8b8b0f"
    )
    print(data)
    result = GlyphParser.parse(data)
    pprint(result)


def decode_metrics():
    data = deHexStr(
        "33 8B 34 8B 35 8B 36 8B 37 8B 38 8B 39 8B 56 FF D2 1F 16 FF 57 8B 3E F9 50 3F 90 40 8B 4B 8B 4C 00 00 08 09 00 00 00 00 00 00 4D FF 00 00 06 68 4E FC 2E 4F F7 52 50 AB 51 FF 00 00 09 C8 52 FF 00 00 05 38 54 8B 8B 32"
    )
    result = MetricsParser.parse(BytesIO(data), len(data))
    pprint(result)


data = "".join(argv[1:])
decode_data(deHexStr(data))
# decode_metrics()



from unittest import TestCase

from vfbLib.ufo.pshints import update_adobe_hinting

ps_hints_empty = """  <hintSetList>
  </hintSetList>"""

ps_hints = """  <hintSetList>
    <hintset pointTag="at01">
      <hstem pos="0" width="50" />
      <hstem pos="358" width="54" />
      <hstem pos="648" width="50" />
      <vstem pos="8" width="123" />
      <vstem pos="227" width="64" />
      <vstem pos="460" width="101" />
      <vstem pos="803" width="106" />
    </hintset>
    <hintset pointTag="sv03">
      <hstem pos="-12" width="51" />
      <hstem pos="-12" width="51" />
      <hstem pos="0" width="50" />
      <hstem pos="358" width="54" />
      <hstem pos="358" width="54" />
      <hstem pos="358" width="54" />
      <hstem pos="648" width="50" />
      <hstem pos="648" width="50" />
      <hstem pos="648" width="50" />
      <vstem pos="8" width="123" />
      <vstem pos="8" width="123" />
      <vstem pos="8" width="123" />
      <vstem pos="227" width="64" />
      <vstem pos="227" width="64" />
      <vstem pos="227" width="64" />
      <vstem pos="460" width="101" />
      <vstem pos="460" width="101" />
      <vstem pos="460" width="101" />
      <vstem pos="803" width="106" />
      <vstem pos="803" width="106" />
      <vstem pos="803" width="106" />
    </hintset>
  </hintSetList>"""

expected_2_hintsets = {
    "hintSetList": [
        {
            "pointTag": "at01",
            "stems": [
                "hstem 0 50",
                "hstem 358 54",
                "hstem 648 50",
                "vstem 8 123",
                "vstem 227 64",
                "vstem 460 101",
                "vstem 803 106",
            ],
        },
        {
            "pointTag": "sv03",
            "stems": [
                "hstem -12 51",
                "hstem -12 51",
                "hstem 0 50",
                "hstem 358 54",
                "hstem 358 54",
                "hstem 358 54",
                "hstem 648 50",
                "hstem 648 50",
                "hstem 648 50",
                "vstem 8 123",
                "vstem 8 123",
                "vstem 8 123",
                "vstem 227 64",
                "vstem 227 64",
                "vstem 227 64",
                "vstem 460 101",
                "vstem 460 101",
                "vstem 460 101",
                "vstem 803 106",
                "vstem 803 106",
                "vstem 803 106",
            ],
        },
    ]
}


class PshintsTest(TestCase):
    def test_empty(self):
        r = update_adobe_hinting(ps_hints_empty)
        assert r == {}

    def test_2_hintsets(self):
        r = update_adobe_hinting(ps_hints)
        assert r == expected_2_hintsets


if __name__ == "__main__":
    t = PshintsTest()
    t.test_2_hintsets()



from unittest import TestCase

from vfbLib.parsers.numeric import (
    Int16Parser,
    # Int32Parser,
    SignedInt16Parser,
    SignedInt32Parser,
)


class Int16ParserTest(TestCase):
    def test_unicode_ranges(self):
        data = "3200"
        expected = 50
        result = Int16Parser().parse_hex(data)
        assert result == expected


class SignedInt16ParserTest(TestCase):
    def test_minus_one(self):
        data = "ffff"
        expected = -1
        result = SignedInt16Parser().parse_hex(data)
        assert result == expected

    def test_hhea_ascender(self):
        data = "ee02"
        expected = 750
        result = SignedInt16Parser().parse_hex(data)
        assert result == expected

    def test_hhea_descender(self):
        data = "06ff"
        expected = -250
        result = SignedInt16Parser().parse_hex(data)
        assert result == expected


class SignedInt32ParserTest(TestCase):
    def test_minus_one(self):
        data = "ffffffff"
        expected = -1
        result = SignedInt32Parser().parse_hex(data)
        assert result == expected



from pathlib import Path
from unittest import TestCase

from vfbLib.ufo.builder import VfbToUfoBuilder


class TestVfb:
    pass


class VfbToUfoBuilderTest(TestCase):
    def test_get_master_path(self):
        ufo_path = Path("/") / "path" / "to" / "ufo" / "my_ufo.ufo"

        # First master without number
        master_path = VfbToUfoBuilder.get_master_path(ufo_path, 0)
        assert master_path == ufo_path

        # Second master with number
        master_path = VfbToUfoBuilder.get_master_path(ufo_path, 1)
        assert master_path == ufo_path.parent / "my_ufo-1.ufo"



from __future__ import annotations

import logging

from vfbLib import export_options, opentype_export_options
from vfbLib.helpers import binaryToIntList
from vfbLib.parsers.base import BaseParser, EncodedKeyValuesParser

logger = logging.getLogger(__name__)


class ExportOptionsParser(BaseParser):
    """
    A parser that reads data a bit field representing export options.
    """

    def _parse(self) -> list[str]:
        val = self.read_uint16()
        bits = binaryToIntList(val)
        options = [export_options.get(i, str(i)) for i in bits]
        return options


class OpenTypeExportOptionsParser(EncodedKeyValuesParser):
    def _parse(self):
        final = []
        for d in super()._parse():
            assert len(d) == 1
            k, v = tuple(d.items())[0]
            final.append({opentype_export_options.get(k, str(k)): v})
        return final



from unittest import TestCase

from vfbLib.ufo.time import convert_timestamp


def fmt(d):
    return d.strftime("%Y/%m/%d %H:%M:%S")


class TestConvertTimestamp(TestCase):
    def test_negative(self):
        ts = 2082841199  # -1
        d = convert_timestamp(ts)
        assert fmt(d) == "1970/01/01 00:00:00"

    def test_origin(self):
        ts = 2082841200  # 01.01.1970, 00:00:00 UTC
        d = convert_timestamp(ts)
        assert fmt(d) == "1970/01/01 00:00:00"

    def test_summertime(self):
        ts = 3303309600
        d = convert_timestamp(ts)
        assert fmt(d) == "2008/09/03 19:00:00"

    def test_now(self):
        ts = 3754301757  # now
        d = convert_timestamp(ts)
        assert fmt(d) == "2022/12/19 14:35:57"



from unittest import TestCase

from vfbLib.compilers.guides import GuidePropertiesCompiler, GuidesCompiler
from vfbLib.parsers.guides import GlobalGuidesParser, GuidePropertiesParser

raw_full = {
    "h": [
        {"index": 1, "color": "#c0c0c0", "name": "wtf"},
        {"index": 2, "color": "#80ff57", "name": "hjk"},
    ],
    "v": [{"index": 2, "color": "#008000", "name": "rsb"}],
}

bin_full = (
    "8c"  # 1: index
    "ff00c0c0c0"  # color
    "8e"  # 3
    "777466"
    "8d"  # 2: index
    "ff0057ff80"  # color
    "8e"  # 3
    "686a6b"
    "8b"  # 0  # end of horizontal guides
    "8d"  # 2: index
    "ff00008000"  # color
    "8e"  # 3
    "727362"
    "8b"  # 0  # end of vertical guides
)

raw_partial = {
    "h": [
        {"index": 1, "color": "#c0c0c0", "name": "wtf"},
        {"index": 2, "color": "#80ff57", "name": "hjk"},
    ],
    "v": [{"index": 2, "name": "rsbwtf"}, {"index": 3, "color": "#ffff00"}],
}

bin_partial = (
    "8c"  # 1: index
    "ff00c0c0c0"  # color
    "8e"  # 3: length of name
    "777466"  # "wtf"
    "8d"  # 2: index
    "ff0057ff80"  # color
    "8e"  # 3: length of name
    "686a6b"  # "hjk"
    "8b"  # 0  # end of horizontal guides
    "8d"  # 2: index
    "8a"  # -1 = no color
    "91"  # 6: length of name
    "727362777466"  # "rsbwtf"
    "8e"  # 3
    "ff0000ffff"  # color
    "8b"  # 0: length of name
    "8b"
)


class GuidePropertiesCompilerTest(TestCase):
    def test_full(self):
        result = GuidePropertiesCompiler().compile_hex(raw_full)
        assert result == bin_full

    def test_partial(self):
        result = GuidePropertiesCompiler().compile_hex(raw_partial)
        assert result == bin_partial


class GuidePropertiesParserTest(TestCase):
    def test_full(self):
        result = GuidePropertiesParser().parse_hex(bin_full)
        assert result == raw_full

    def test_partial(self):
        result = GuidePropertiesParser().parse_hex(bin_partial)
        assert result == raw_partial


bin_guides = (
    "9e"  # 19 horizontal guides
    "fa688b"  # 980 0 (master 0)
    "fa688b"  # 980 0 (master 1)
    "fbd48b"  # -320 0 (master 0)
    "fbd48b"  # -320 0 (master 1)
    "f7be8b"
    "f7be8b"
    "f9ca8b"
    "f9cb8b"
    "f9c98b"
    "f9c38b"
    "f9378b"
    "f9378b"
    "f93f8b"
    "f9418b"
    "f8ba8b"
    "f8b78b"
    "f7f48b"
    "f7ec8b"
    "f7ec8b"
    "f7e28b"
    "f7528b"
    "f7528b"
    "f74d8b"
    "f74b8b"
    "388b"
    "388b"
    "fb238b"
    "fb108b"
    "fb2a8b"
    "fb168b"
    "fb348b"
    "fb1d8b"
    "f7c78b"
    "f7c78b"
    "f9b48b"
    "f9b68b"
    "f9ac8b"
    "f9ac8b"
    "8c"  # 1 vertical guide
    "fbb68b"  # -290 0 (master 0)
    "fbae8b"  # -282 0 (master 1)
)

raw_guides = {
    "h": [
        [
            {"pos": 980, "angle": 0.0},
            {"pos": -320, "angle": 0.0},
            {"pos": 298, "angle": 0.0},
            {"pos": 822, "angle": 0.0},
            {"pos": 821, "angle": 0.0},
            {"pos": 675, "angle": 0.0},
            {"pos": 683, "angle": 0.0},
            {"pos": 550, "angle": 0.0},
            {"pos": 352, "angle": 0.0},
            {"pos": 344, "angle": 0.0},
            {"pos": 190, "angle": 0.0},
            {"pos": 185, "angle": 0.0},
            {"pos": -83, "angle": 0.0},
            {"pos": -143, "angle": 0.0},
            {"pos": -150, "angle": 0.0},
            {"pos": -160, "angle": 0.0},
            {"pos": 307, "angle": 0.0},
            {"pos": 800, "angle": 0.0},
            {"pos": 792, "angle": 0.0},
        ],
        [
            {"pos": 980, "angle": 0.0},
            {"pos": -320, "angle": 0.0},
            {"pos": 298, "angle": 0.0},
            {"pos": 823, "angle": 0.0},
            {"pos": 815, "angle": 0.0},
            {"pos": 675, "angle": 0.0},
            {"pos": 685, "angle": 0.0},
            {"pos": 547, "angle": 0.0},
            {"pos": 344, "angle": 0.0},
            {"pos": 334, "angle": 0.0},
            {"pos": 190, "angle": 0.0},
            {"pos": 183, "angle": 0.0},
            {"pos": -83, "angle": 0.0},
            {"pos": -124, "angle": 0.0},
            {"pos": -130, "angle": 0.0},
            {"pos": -137, "angle": 0.0},
            {"pos": 307, "angle": 0.0},
            {"pos": 802, "angle": 0.0},
            {"pos": 792, "angle": 0.0},
        ],
    ],
    "v": [[{"pos": -290, "angle": 0.0}], [{"pos": -282, "angle": 0.0}]],
}

bin_guides_empty = "8b8b"

raw_guides_empty = {"h": [[], []], "v": [[], []]}


class GlobalGuidesParserTest(TestCase):
    def test(self):
        result = GlobalGuidesParser().parse_hex(bin_guides, master_count=2)
        assert result == raw_guides

    def test_empty(self):
        result = GlobalGuidesParser().parse_hex(bin_guides_empty, master_count=2)
        assert result == raw_guides_empty


class GuidesCompilerTest(TestCase):
    def test(self):
        result = GuidesCompiler().compile_hex(raw_guides, master_count=2)
        assert result == bin_guides

    def test_empty(self):
        result = GuidesCompiler().compile_hex(raw_guides_empty, master_count=2)
        assert result == bin_guides_empty



from pathlib import Path
from unittest import TestCase

from fontTools.designspaceLib import DesignSpaceDocument

from vfbLib.ufo.builder import VfbToUfoBuilder
from vfbLib.vfb.vfb import Vfb


def data_path():
    return Path(__file__).parent / "Data"


def vfb_path(name: str) -> Path:
    return data_path() / name


class GlyphCompilerTest(TestCase):
    def test_vfb_to_ufo_mm(self):
        vfb = Vfb(vfb_path("masters.vfb"))
        vfb.decompile()
        builder = VfbToUfoBuilder(vfb, move_groups=False)
        ufos, designspace = builder.get_ufos_designspace(data_path())
        assert len(ufos) == 4
        assert isinstance(designspace, DesignSpaceDocument)

        # Master 0
        ufo = ufos[0]
        assert len(ufo) == 2  # 1 glyph only
        glyph = ufo["a"]
        assert len(glyph) == 1
        assert glyph.width == 353
        point = glyph.contours[0].points[3]
        assert (point.x, point.y) == (176, 181)

        # Master 1
        ufo = ufos[1]
        glyph = ufo["a"]
        assert glyph.width == 457
        point = glyph.contours[0].points[3]
        assert (point.x, point.y) == (228, 161)

        # Master 2
        ufo = ufos[2]
        glyph = ufo["a"]
        assert glyph.width == 437
        point = glyph.contours[0].points[3]
        assert (point.x, point.y) == (218, 181)

        # Master 3
        ufo = ufos[3]
        glyph = ufo["a"]
        assert glyph.width == 565
        point = glyph.contours[0].points[3]
        assert (point.x, point.y) == (281, 151)

    def test_vfb_to_ufo_mm_reader_deepcopy(self):
        # Check that objects have been copied and are not shared anymore between UFOs
        vfb = Vfb(vfb_path("masters.vfb"))
        vfb.decompile()
        builder = VfbToUfoBuilder(vfb, move_groups=False)
        ufos = builder.get_ufo_masters(silent=True)
        assert len(ufos) == 4

        ufo0, ufo1, _, _ = ufos

        # Features

        fea0 = (
            "\n\nlanguagesystem DFLT dflt;\nfeature kern {\n  pos a t 100 ;\n} kern;\n"
        )

        assert ufo0.features.text == fea0
        assert ufo1.features.text == fea0

        # Glyphs

        glyph0 = ufo0["a"]
        assert glyph0.width == 353
        glyph0.width = 354
        point = glyph0.contours[0].points[3]
        assert (point.x, point.y) == (176, 181)
        point.x = 200
        glyph0.lib["foo"] = "bar"
        assert glyph0.unicode == 0x61
        glyph0.unicodes.append(0x41)

        # Master 1
        glyph1 = ufo1["a"]
        assert glyph1.width == 457
        point = glyph1.contours[0].points[3]
        assert (point.x, point.y) == (228, 161)
        assert "foo" not in glyph1.lib
        assert glyph1.unicodes == [0x61]

        # Groups

        expected_group_order = [
            "public.kern1._a_LEFT",
            "public.kern2._a_RIGHT",
            "public.kern1._t",
            "public.kern2._t",
            # "_empty",
            ".mtrx5",
            "letters",
        ]

        expected_groups = dict(
            zip(
                expected_group_order,
                [
                    ["a"],
                    ["a"],
                    ["t"],
                    ["t"],
                    # [],
                    ["a"],
                    ["a", "t"],
                ],
            )
        )

        assert ufo0.groups == expected_groups
        assert ufo0.groups == ufo1.groups
        del ufo1.groups[".mtrx5"]
        assert ".mtrx5" in ufo0.groups
        ufo0.groups["_empty"] = ["a"]
        assert "_empty" not in ufo1.groups

        # Kerning

        ufo0.kerning[("public.kern1._a_LEFT", "public.kern2._a_RIGHT")] = -10
        assert ("public.kern1._a_LEFT", "public.kern2._a_RIGHT") not in ufo1.kerning
        assert ufo0.kerning[("public.kern1._a_LEFT", "public.kern2._t")] == -73
        ufo0.kerning[("public.kern1._a_LEFT", "public.kern2._t")] = -72
        assert ufo0.kerning[("public.kern1._a_LEFT", "public.kern2._t")] == -72
        assert ufo1.kerning[("public.kern1._a_LEFT", "public.kern2._t")] == -166

        # Lib

        assert ufo0.lib["public.groupOrder"] == expected_group_order
        assert "com.lucasfonts.vfblib" in ufo0.lib["com.fontlab.v5.userData"]
        del ufo0.lib["com.fontlab.v5.userData"]
        assert "com.lucasfonts.vfblib" in ufo1.lib["com.fontlab.v5.userData"]
        ufo0.lib["foo"] = "bar"
        assert "foo" not in ufo1.lib



from unittest import TestCase

from vfbLib.ufo.kerning import UfoKerning

glyph_order = ["A", "Adieresis", "E", "F", "T", "Tbar", "Y"]

groups = {
    "public.kern1.A": ["A", "Adieresis"],
    "public.kern2.A": ["A", "Adieresis"],
    "public.kern1.T": ["T", "Y"],
    "public.kern2.T": ["T", "Y", "Tbar"],
    "public.kern1.F": ["F", "Tbar"],
    "public.kern2.E": ["E", "F"],
}

mmk = {}

kg = {
    "public.kern1.A": "A",
    "public.kern2.A": "A",
    "public.kern1.T": "T",
    "public.kern2.T": "T",
    "public.kern1.F": "F",
    "public.kern2.E": "E",
}

kerning = UfoKerning(
    glyph_order=glyph_order, groups=groups, mm_kerning=mmk, key_glyphs=kg
)
kerning.mm_kerning_names = {
    ("public.kern1.A", "public.kern2.T"): [-100, -80],
    ("public.kern1.T", "public.kern2.A"): [-100, -80],
    ("Adieresis", "public.kern2.T"): [-80, -60],
    ("public.kern1.T", "Adieresis"): [-80, -60],
    ("public.kern1.F", "public.kern2.A"): [-70, -65],
    ("Tbar", "Adieresis"): [-60, -55],
    ("P", "A"): [-90, -85],
    ("P", "J"): [-90, -85],
}


class UfoKerningTest(TestCase):
    def test_exception_keypair(self):
        assert not kerning._is_exception("A", "T")

    def test_exception_key_member(self):
        assert kerning._is_exception("T", "Adieresis")

    def test_exception_member_key(self):
        assert kerning._is_exception("Adieresis", "T")

    def test_exception_member_member(self):
        assert kerning._is_exception("Tbar", "Adieresis")

    def test_exception_nonmember_key(self):
        assert not kerning._is_exception("P", "A")

    def test_exception_key_nonmember(self):
        assert not kerning._is_exception("T", "J")

    def test_exception_nonmember_member(self):
        assert kerning._is_exception("Ydieresis", "Adieresis")

    def test_exception_member_nonmember(self):
        assert kerning._is_exception("Adieresis", "Ydieresis")

    def test_exception_nonmember_nonmember(self):
        assert not kerning._is_exception("P", "J")

    def test_exception_unkerned(self):
        assert not kerning._is_exception("X", "hyphen")



from pathlib import Path

from vfbLib.vfb.vfb import Vfb

size(1000, 1000)

vfb_path = Path(__file__).parent.parent / "Tests" / "Data" / "ComicJensPro-Regular.vfb"

vfb = Vfb(vfb_path)
g = vfb["Adieresis"]
bez = BezierPath()
g.drawPoints(bez)
drawPath(bez)



from __future__ import annotations

from vfbLib import export_options, opentype_export_options
from vfbLib.compilers.base import BaseCompiler
from vfbLib.helpers import intListToBinary


class ExportOptionsCompiler(BaseCompiler):
    def _compile(self, data: list[str]) -> None:
        rev = {v: k for k, v in export_options.items()}
        options = [rev[k] for k in data]
        self.write_uint16(intListToBinary(options))


class OpenTypeExportOptionsCompiler(BaseCompiler):
    def _compile(self, data: list[dict[str, int]]) -> None:
        rev = {v: k for k, v in opentype_export_options.items()}
        for opt in data:
            raw_key, v = tuple(opt.items())[0]
            if raw_key in rev:
                k = rev[raw_key]
            else:
                k = int(raw_key)
            self.write_uint8(k)
            self.write_value(v)
        self.write_uint8(0x64)



from unittest import TestCase

from vfbLib.parsers.pclt import PcltParser


class EncodedValueListParserTest(TestCase):
    def test_pclt_empty(self):
        data = (
            "8b8b8b8b8b8b8b"
            "00000000000000"
            "00000000000000"
            "00000000000000"
            "00000000000000"
            "0000000000"
        )

        expected = {
            "font_number": 0,
            "pitch": 0,
            "x_height": 0,
            "style": 0,
            "type_family": 0,
            "cap_height": 0,
            "symbol_set": 0,
            "typeface": (
                "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            ),
            "character_complement": [
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
            ],
            "file_name": "\x00\x00\x00\x00\x00\x00",
            "stroke_weight": 0,
            "width_type": 0,
            "serif_style": 0,
        }
        result = PcltParser().parse_hex(data)
        assert result == expected

    def test_pclt(self):
        data = (
            "ffcd000003c2f88ad4ff00004002f952a4"
            "534d4820202020202020202020202020"
            "1a34567809123400"
            "544e5252300002fe82"
        )
        expected = {
            "font_number": 3439329283,
            "pitch": 55,
            "x_height": 502,
            "style": 73,
            "type_family": 16386,
            "cap_height": 702,
            "symbol_set": 25,
            "typeface": "SMH             ",
            "character_complement": [
                0x1A,
                0x34,
                0x56,
                0x78,
                0x09,
                0x12,
                0x34,
                0x00,
            ],
            "file_name": "TNRR0\x00",
            "stroke_weight": 2,
            "width_type": -2,
            "serif_style": 130,
        }
        result = PcltParser().parse_hex(data)
        assert result == expected



from io import BytesIO
from unittest import TestCase

from fontTools.misc.textTools import hexStr

from vfbLib.compilers.value import write_value, write_value_long


class ValueEncoderTest(TestCase):
    def expect(self, encoded, decoded, signed=True):
        data = BytesIO()
        write_value(decoded, data, signed)
        assert hexStr(data.getvalue()) == encoded

    def expect_unoptimized(self, encoded, decoded, signed=True):
        data = BytesIO()
        write_value_long(decoded, data, signed)
        assert hexStr(data.getvalue()) == encoded

    def test_0x20(self):
        self.expect("20", -107)

    def test_0x8b(self):
        self.expect("8b", 0)

    def test_0xf6(self):
        self.expect("f6", 107)

    def test_0xf700(self):
        self.expect("f700", 108)

    def test_0xf701(self):
        self.expect("f701", 109)

    def test_0xf7ff(self):
        self.expect("f7ff", 363)

    def test_0xf800(self):
        self.expect("f800", 364)

    def test_0xf801(self):
        self.expect("f801", 365)

    def test_0xfa00(self):
        self.expect("fa00", 876)

    def test_0xfaff(self):
        self.expect("faff", 1131)

    def test_0xfb00(self):
        # 64256
        self.expect("fb00", -108)

    def test_0xfb01(self):
        self.expect("fb01", -109)

    def test_0xfe00(self):
        self.expect("fe00", -876)

    def test_0xfeff(self):
        self.expect("feff", -1131)

    def test_0xff00000000(self):
        self.expect_unoptimized("ff00000000", 0)

    def test_0xff00001000(self):
        self.expect("ff00001000", 4096)

    def test_0xff00001000u(self):
        self.expect("ff00001000", 4096, False)

    def test_0xffffffffff(self):
        self.expect_unoptimized("ffffffffff", -1)

    def test_0xffffffffffu(self):
        # 4294967295 (max)
        self.expect("ffffffffff", 0xFFFFFFFF, False)

    def test_0xffffffefff(self):
        self.expect("ffffffefff", -4097)

    def test_0xffffffefffu(self):
        # 4294963199
        self.expect("ffffffefff", 0xFFFFEFFF, False)



from unittest import TestCase

from vfbLib.parsers.text import (
    StringParser,
)


class StringParserTest(TestCase):
    def test_links_1(self):
        data = "57 74 30 20 57 64 31 20"
        expected = "Wt0 Wd1 "  # Trailing space does not get stripped
        result = StringParser().parse_hex(data)
        assert result == expected



from unittest import TestCase

from vfbLib.helpers import binaryToIntList, intListToBinary


class HelpersTest(TestCase):
    def test_binaryToIntList0(self):
        assert binaryToIntList(0) == []

    def test_binaryToIntList1(self):
        assert binaryToIntList(1) == [0]

    def test_binaryToIntList2(self):
        assert binaryToIntList(2) == [1]

    def test_binaryToIntList3(self):
        assert binaryToIntList(3) == [0, 1]

    def test_intListToBinary0(self):
        assert intListToBinary([]) == 0

    def test_intListToBinary1(self):
        assert intListToBinary([0]) == 1

    def test_intListToBinary2(self):
        assert intListToBinary([1]) == 2

    def test_intListToBinary3(self):
        assert intListToBinary([0, 1]) == 3



from __future__ import annotations

import logging
from enum import Enum
from struct import unpack
from typing import TYPE_CHECKING, Any

from fontTools.ttLib.tables.ttProgram import Program

from vfbLib import DIRECTIONS, gdef_class_names, replace_types
from vfbLib.helpers import hexStr
from vfbLib.parsers.base import BaseParser
from vfbLib.parsers.guides import parse_guides
from vfbLib.truetype import TT_COMMANDS
from vfbLib.typing import (
    AnchorDict,
    Component,
    GdefDict,
    GlyphData,
    HintDict,
    Instruction,
    LinkDict,
    MaskData,
    MMAnchorDict,
    MMHintsDict,
    MMNode,
    Point,
)
from vfbLib.value import read_value

if TYPE_CHECKING:
    from io import BytesIO


logger = logging.getLogger(__name__)


class PathCommand(Enum):
    move = 0
    line = 1
    curve = 3
    qcurve = 4


def read_absolute_point(
    points: list[list[Any]],
    stream: BytesIO,
    num_masters: int,
    x_masters: list[int],
    y_masters: list[int],
):
    # Read coordinates for all masters from stream, add the points to a list
    # of points per master. Keep track of the relative coordinates reference
    # value in x_masters and y_masters.
    for m in range(num_masters):
        # End point
        xrel = read_value(stream)
        yrel = read_value(stream)
        x_masters[m] += xrel
        y_masters[m] += yrel
        points[m].append((x_masters[m], y_masters[m]))


class GlyphAnchorsParser(BaseParser):
    def _parse(self) -> list[MMAnchorDict]:
        anchors = []
        num_anchors = self.read_value()
        num_masters = self.read_value()
        for _ in range(num_anchors):
            anchor = MMAnchorDict(x=[], y=[])
            for _ in range(num_masters):
                anchor["x"].append(self.read_value())
                anchor["y"].append(self.read_value())
            anchors.append(anchor)
        return anchors


class GlyphAnchorsSuppParser(BaseParser):
    def _parse(self) -> list:
        anchors = []
        num_anchors = self.read_value()
        for _ in range(num_anchors):
            hue = self.read_value()
            rv1 = self.read_value()
            anchors.append({"hue": hue, "reserved": rv1})
        return anchors


class GlyphGDEFParser(BaseParser):
    def _parse(self) -> GdefDict:
        gdef: GdefDict = {}
        glyph_class = self.read_value()
        glyph_class_name = gdef_class_names[glyph_class]
        if glyph_class_name != "unassigned":
            gdef["glyph_class"] = glyph_class_name

        num_anchors = self.read_value()
        anchors = []
        for _ in range(num_anchors):
            name = self.read_str_with_len()
            x = self.read_value()
            x1 = self.read_value()
            y = self.read_value()
            y1 = self.read_value()
            anchor = AnchorDict(x=x, x1=x1, y=y, y1=y1)
            if name:
                anchor["name"] = name
            anchors.append(anchor)
        if anchors:
            gdef["anchors"] = anchors

        num_carets = self.read_value()
        carets = []
        for _ in range(num_carets):
            pos = self.read_value()
            xxx = self.read_value()
            carets.append((pos, xxx))
        if carets:
            gdef["carets"] = carets

        try:
            num_values = self.read_value()
            values = [self.read_value() for _ in range(num_values)]
            if values:
                gdef["unknown"] = values
        except EOFError:
            pass

        return gdef


class GlyphOriginParser(BaseParser):
    def _parse(self) -> dict[str, Any]:
        x = self.read_int16()
        y = self.read_int16()
        return {"x": x, "y": y}


class GlyphParser(BaseParser):
    def parse_guides(self) -> None:
        # Guidelines
        guides = parse_guides(self.stream, self.master_count, f"glyph '{self.name}'")
        if guides:
            self.glyphdata["guides"] = guides

    def parse_binary(self) -> None:
        # Imported binary TrueType data
        imported: dict[str, Any] = {}
        while True:
            key = self.read_uint8()

            if key == 0x28:
                break

            elif key == 0x29:
                # Metrics
                imported["width"] = self.read_value()
                imported["lsb"] = self.read_value()
                imported["unknown1"] = self.read_value()
                imported["unknown2"] = self.read_value()
                imported["unknown3"] = self.read_value()
                imported["bbox"] = [self.read_value() for _ in range(4)]

            elif key == 0x2A:
                # Outlines
                num_contours = self.read_value()
                imported["endpoints"] = [self.read_value() for _ in range(num_contours)]
                num_nodes = self.read_value()
                nodes = []
                x = 0
                y = 0
                for _ in range(num_nodes):
                    x += self.read_value()
                    y += self.read_value()
                    byte = self.read_uint8()
                    flags = byte >> 4
                    cmd = byte & 0x0F
                    node = (hex(cmd), hex(flags), x, y)
                    nodes.append(node)
                if nodes:
                    imported["nodes"] = nodes

            elif key == 0x2B:
                # Instructions
                num_instructions = self.read_value()
                instructions = self.stream.read(num_instructions)
                p = Program()
                p.fromBytecode(instructions)
                imported["instructions"] = p.getAssembly()

            elif key == 0x2C:
                # Probably HDMX data
                num = self.read_value()
                imported["hdmx"] = [self.read_uint8() for _ in range(num)]

            else:
                logger.error(imported)
                logger.error(f"Unknown key in imported binary glyph data: {hex(key)}")
                raise ValueError

        self.glyphdata["imported"] = imported

    def parse_components(self) -> None:
        components = []
        num = self.read_value()
        for i in range(num):
            gid = self.read_value()
            c = Component(gid=gid, offsetX=[], offsetY=[], scaleX=[], scaleY=[])
            for _ in range(self.master_count):
                x = self.read_value()
                y = self.read_value()
                scaleX, scaleY = self.read_doubles(2)
                c["offsetX"].append(x)
                c["offsetY"].append(y)
                c["scaleX"].append(scaleX)
                c["scaleY"].append(scaleY)
            components.append(c)
        self.glyphdata["components"] = components

    def parse_hints(self) -> None:
        hints = MMHintsDict(v=[], h=[])
        for direction in DIRECTIONS:
            num_hints = self.read_value()
            for _ in range(num_hints):
                master_hints = []
                for _ in range(self.master_count):
                    pos = self.read_value()
                    width = self.read_value()
                    master_hints.append(HintDict(pos=pos, width=width))
                hints[direction].append(master_hints)

        num_hintmasks = self.read_value()
        if num_hintmasks > 0:
            hintmasks: list[tuple[str, int]] = []
            for _ in range(num_hintmasks):
                k = self.read_uint8()
                val = self.read_value()
                key = replace_types[k]
                hintmasks.append((key, val))
            if hintmasks:
                hints["hintmasks"] = hintmasks

        if hints["v"] or hints["h"] or "hintmasks" in hints:
            self.glyphdata["hints"] = hints

    def parse_instructions(self) -> None:
        # Number of bytes for instructions that follow;
        # we don't use it
        _ = self.read_value()
        num_commands = self.read_value()
        commands: list[Instruction] = []
        for i in range(num_commands):
            cmd = self.read_uint8()
            params = [self.read_value() for _ in range(len(TT_COMMANDS[cmd]["params"]))]
            commands.append(
                Instruction(
                    cmd=TT_COMMANDS[cmd]["name"],
                    params=dict(zip(TT_COMMANDS[cmd]["params"], params)),
                )
            )
        # Instructions are ended by 3 * 0!?
        for _ in range(3):
            self.read_value()

        if commands:
            self.glyphdata["tth"] = commands

    def parse_metrics(self) -> None:
        metrics: list[Point] = []
        for _ in range(self.master_count):
            master_metrics = (
                self.read_value(),
                self.read_value(),
            )
            metrics.append(master_metrics)
        self.glyphdata["metrics"] = metrics

    def parse_outlines(self, target: GlyphData | MaskData) -> int:
        # Nodes
        master_count = self.read_value()
        target["num_masters"] = master_count

        # 2 x the number of values to be read after num_nodes, the reason is unclear.
        _ = self.read_value()
        # glyphdata["num_node_values"] = num_node_values

        num_nodes = self.read_value()
        segments: list[MMNode] = []
        x = [0 for _ in range(master_count)]
        y = [0 for _ in range(master_count)]

        for _ in range(num_nodes):
            byte = self.read_uint8()
            flags = byte >> 4
            cmd = PathCommand(byte & 0x0F).name

            # End point
            points: list[list[Point]] = [[] for _ in range(master_count)]
            read_absolute_point(points, self.stream, master_count, x, y)

            if cmd == "curve":
                # First control point
                read_absolute_point(points, self.stream, master_count, x, y)
                # Second control point
                read_absolute_point(points, self.stream, master_count, x, y)

            segment = MMNode(type=cmd, flags=flags, points=points)
            segments.append(segment)
        target["nodes"] = segments
        return master_count

    def parse_kerning(self) -> None:
        num = self.read_value()
        kerning = {}
        for _ in range(num):
            # Glyph index of right kerning partner
            gid = self.read_value()
            values = []
            for _ in range(self.master_count):
                # List of values, one value per master
                values.append(self.read_value())
            kerning[gid] = values
        self.glyphdata["kerning"] = kerning

    def _parse(self) -> dict[str, Any]:
        """
        01090701
        01  92[7]2e 6e 6f 74 64 65 66 # Glyph name
        08  8c bb[48] 93[8]
            00 ab 8b    [  32,    0] move
            01 8b f9a0  [   0,  780] line 32,780
            01 f7fc 8b  [ 360,    0] line 392,780
            01 8b fda0  [   0, -780] line 392, 0
            00 fbcc bb  [-312,   48] move  80, 48
            01 f79c 8b  [ 264,    0] line
            01 8b f940  [   0,  684] line
            01 fb9c 8b  [-264,    0] line
        02  f83c 8b  [ 424,    0]
        03  8b 8b 8b
        04  8b 8b
        0a  a1[22] 8f[4] # TT hinting
            02 8b 8b
            04 8b 8c 89 8a
            04 8b 8f 89 8a
            04 8c 92 89 8a
            8b 8b 8b
        0f
        """
        self.glyphdata = GlyphData()
        start = unpack("<4B", self.stream.read(4))
        if start != (1, 9, 7, 1):
            logger.warning(
                f"Unexpected glyph constant: {start}, please notify developer"
            )
        # glyphdata["constants"] = start
        while True:
            # Read a value to decide what kind of information follows
            v = self.read_uint8()

            if v == 0x01:
                # Glyph name
                glyph_name = self.read_str_with_len()
                self.glyphdata["name"] = glyph_name
                self.name = self.glyphdata["name"]

            elif v == 0x02:
                # Metrics
                self.parse_metrics()

            elif v == 0x03:
                # PS Hints
                self.parse_hints()

            elif v == 0x04:
                # Guides
                self.parse_guides()

            elif v == 0x05:
                # Components
                self.parse_components()

            elif v == 0x06:
                # Kerning
                self.parse_kerning()

            elif v == 0x08:
                # Outlines
                self.master_count = self.parse_outlines(self.glyphdata)

            elif v == 0x09:
                # Imported binary TrueType data
                self.parse_binary()

            elif v == 0x0A:
                # TrueType instructions
                self.parse_instructions()

            elif v == 0x0F:
                # logger.debug("Glyph done.")
                break

            else:
                logger.error(f"Unhandled info field: {hex(v)}")
                logger.error(hexStr(self.stream.read()))
                raise ValueError

        return dict(self.glyphdata)


class GlyphUnicodeParser(BaseParser):
    def _parse(self) -> list:
        unicodes = []
        for _ in range(self.stream.getbuffer().nbytes // 2):
            u = self.read_uint16()
            unicodes.append(u)
        return unicodes


class GlyphUnicodeSuppParser(BaseParser):
    def _parse(self) -> list:
        unicodes = []
        for _ in range(self.stream.getbuffer().nbytes // 4):
            u = self.read_uint32()
            unicodes.append(u)
        return unicodes


class LinkParser(BaseParser):
    def _parse(self) -> LinkDict:
        links = LinkDict(x=[], y=[])
        for i in range(2):
            num = self.read_value()
            for _ in range(num):
                src = self.read_value()
                tgt = self.read_value()
                links[("y", "x")[i]].append((src, tgt))
        return links


class MaskParser(GlyphParser):
    def _parse(self) -> dict[str, Any]:
        num = self.read_value()
        maskdata = MaskData(num=num)
        for i in range(num):
            maskdata[f"reserved{i}"] = self.read_value()

        # From here, the mask is equal to the outlines
        self.parse_outlines(maskdata)
        return dict(maskdata)


class GlobalMaskParser(GlyphParser):
    def _parse(self) -> dict[str, Any]:
        maskdata = GlyphData()
        self.parse_outlines(maskdata)
        return dict(maskdata)


class MaskMetricsParser(BaseParser):
    # advance width/height for master 0
    def _parse(self) -> tuple[int, int]:
        # FIXME: Normally, the values are little endian.
        # In some cases, the value seems to be saved as big endian, though
        # Maybe try to interpret the value as BE when it is negative and less than a
        # certain sensible value.
        x = self.read_int16()
        y = self.read_int16()
        return (x, y)


class MaskMetricsMMParser(BaseParser):
    # advance width/height for master 1 to 15
    def _parse(self) -> list[tuple[int, int]]:
        values = []
        for _ in range(self.master_count - 1):
            x = self.read_value()
            y = self.read_value()
            values.append((x, y))
        return values


class GlyphSketchParser(BaseParser):
    def _parse(self) -> list[tuple[int, int, int]]:
        num = self.read_value()
        return [
            (self.read_value(), self.read_value(), self.read_value())
            for _ in range(num)
        ]



from unittest import TestCase

from vfbLib.compilers.base import (
    OpenTypeKerningClassFlagsCompiler,
    OpenTypeMetricsClassFlagsCompiler,
)


class OpenTypeKerningClassFlagsCompilerTest(TestCase):
    def test(self):
        expected = (
            "8e"  # 3: number of entries
            "92"  # 7: string length
            "5f615f4c454654"
            "fa94"  # 1024
            "8b"  # 0
            "93"  # 8
            "5f615f5249474854"
            "ff00000800"
            "8b"  # 0
            "8d"  # 2
            "5f74"
            "ff00000c00"
            "8b"  # 0
        )
        result = OpenTypeKerningClassFlagsCompiler().compile_hex(
            {"_a_LEFT": [1024, 0], "_a_RIGHT": [2048, 0], "_t": [3072, 0]}
        )
        assert result == expected


class OpenTypeMetricsClassFlagsCompilerTest(TestCase):
    def test(self):
        expected = (
            "8c"  # 1: number of entries
            "91"  # 6: string length
            "2e6d74727835"
            "8b"  # 0
            "ff00001401"
            "8b"  # 0
        )
        result = OpenTypeMetricsClassFlagsCompiler().compile_hex(
            {".mtrx5": [0, 5121, 0]}
        )
        assert result == expected



from __future__ import annotations

import logging

from vfbLib.parsers.base import BaseParser

logger = logging.getLogger(__name__)


class FL3Type1410Parser(BaseParser):
    """
    A parser that reads the special-cased 1410 entry, which is apparently only
    written by FontLab 3.
    """

    def _parse(self) -> list[int]:
        values = []
        for _ in range(5):
            values.append(self.read_int16())
        return values



from __future__ import annotations

import logging
from io import BufferedReader, BytesIO
from typing import Any

from vfbLib.parsers.base import StreamReader

logger = logging.getLogger(__name__)


class VfbHeaderParser(StreamReader):
    def __init__(self, stream: BufferedReader) -> None:
        self.encoding = "cp1252"
        self.stream: BufferedReader = stream  # type: ignore

    def parse(self) -> dict[str, Any]:
        header: dict[str, Any] = {}
        header["header0"] = self.read_uint8()
        header["filetype"] = self.read_str(5)
        header["header1"] = self.read_uint16()
        size = self.read_uint16()
        chunk1 = []
        for _ in range(size):
            chunk1.append(self.read_uint8())
        header["chunk1"] = chunk1
        if chunk1[-2:] == [10, 0]:
            # FL4+ additions over FL3
            size = self.read_uint16()
            fl4_chunk = self.stream.read(size)
            sr = StreamReader()
            sr.stream = BytesIO(fl4_chunk)
            app_info = {}
            while True:
                key = sr.read_uint8()
                if key == 0:
                    break
                value = sr.read_value()
                if key == 2:
                    # app version
                    app_version = []
                    for i in (24, 16, 8, 0):
                        # Extract the bytes from the number again
                        app_version.append(value >> i & 0xFF)
                    app_info[str(key)] = app_version
                else:
                    app_info[str(key)] = value
            header["creator"] = app_info
            # Two bytes follow that are at the end of chunk1 in the older format
            header["end0"] = self.read_uint8()  # 6
            header["end1"] = self.read_uint8()  # 1
        else:
            # TODO: Just upgrade the header to the newer format?
            # chunk1[-2:] = [10, 0]
            # header["creator"] = {
            #     "1": 1,
            #     "2": [5, 2, 2, 128],
            #     "3": 0,
            # }
            # header["end0"] = 6
            # header["end1"] = 1
            pass
        # The common 0 word at the end of the header
        header["end2"] = self.read_uint16()

        return header



from __future__ import annotations

from pathlib import Path

import orjson

from vfbLib.vfb.vfb import Vfb


def save_vfb_json(
    vfb_path: Path,
    out_path: Path | None = None,
    only_header: bool = False,
    minimal: bool = False,
    unicode_strings: bool = False,
    no_decompile: bool = False,
) -> None:
    vfb = Vfb(
        vfb_path,
        only_header=only_header,
        minimal=minimal,
        unicode_strings=unicode_strings,
    )
    if not no_decompile:
        vfb.decompile()
    suffix = ".vfb.json"
    if out_path:
        out_path = (Path(out_path) / vfb_path.name).with_suffix(suffix)
    else:
        out_path = vfb_path.with_suffix(suffix)
    with open(str(out_path), "wb") as f:
        f.write(
            orjson.dumps(
                vfb.as_dict(),
                option=orjson.OPT_INDENT_2 | orjson.OPT_NON_STR_KEYS,
            )
        )



from __future__ import annotations

from typing import Any

from vfbLib.compilers.base import BaseCompiler


class AnisotropicInterpolationsCompiler(BaseCompiler):
    # FIXME: Untested
    def _compile(self, data: Any) -> None:
        for axis_mappings in data:
            self.write_value(len(axis_mappings))
            for src, tgt in axis_mappings:
                self.write_value(src)
                self.write_value(tgt)


class AxisMappingsCountCompiler(BaseCompiler):
    def _compile(self, data: Any) -> None:
        for value in data:
            self.write_uint32(value)


class AxisMappingsCompiler(BaseCompiler):
    def _compile(self, data: Any) -> None:
        for src, tgt in data:
            self.write_double(src)
            self.write_double(tgt)


class MasterLocationCompiler(BaseCompiler):
    def _compile(self, data: Any) -> None:
        master_index, location = data
        self.write_uint32(master_index)
        self.write_doubles(location)


class PrimaryInstancesCompiler(BaseCompiler):
    def _compile(self, data: Any) -> None:
        # FIXME: Untested
        self.write_value(len(data))
        for instance in data:
            name = instance["name"]
            self.write_str_with_len(name)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Literal

# Used by guides and links
DIRECTIONS: Sequence[Literal["h", "v"]] = ("h", "v")

GLYPH_CONSTANT = (1, 9, 7, 1)

export_options = {
    0: "use_custom_opentype_export_options",
    # 1: "use_default_opentype_export_options",
    2: "use_custom_cmap_encoding",
}

gdef_class_names = (
    "unassigned",  # 0
    "base",  # 1
    "ligature",  # 2
    "mark",  # 3
    "component",  # 4
)

replace_types = {
    0x01: "h",  # hintmask for hstem
    0x02: "v",  # hintmask for vstem
    0xFF: "r",  # Replacement point
    # FIXME: This seems to be the node index of the replacement
    # point. But sometimes it is negative, why?
}

mapping_modes = {
    0: "names_or_index",
    1: "unicode_ranges",
    3: "codepages",
}

opentype_export_options = {
    16: "opentype_name_records",
    #    0 - "append_opentype_records_to_default_names"
    #    1 - "do_not_export_opentype_name_records"
    #    2 - "export_only_opentype_name_records"
    17: "codepage_for_cmap_1_0",
    #    0 - MacRoman (default)
    #    1 - Current codepage in font window
    #    2 - Windows 1250 Central European
    # Didn't bother to check from here on:
    #    3 - Windows 1251 Cyrillic
    #    4 - Windows 1252 Western (ANSI)
    #    ...
    #  153 - NeXT OS NextStep Multinational
    18: "dont_ignore_unicode_indexes",
    20: "autohinting_options",
    #   bit  1 - 7: "single_link_attachment_precision"
    #   bit  8 - "generate_triple_hints"
    #   bit  9 - "generate_delta_instructions"
    #   bit 10 - "direct_links_to_center_of_the_glyph_where_possible"
    #   bit 11 - "interpolate_positions_of_cusp_points"
    #   bit 12 - "interpolate_positions_of_double_links"
    #   bit 14 - "add_link_to_rsb"
    21: "export_hinted_truetype_font",
    22: "autohint_unhinted_glyphs",
    23: "keep_existing_truetype_instructions",
    24: "export_visual_truetype_hints",
    28: "export_embedded_bitmaps",
    29: "copy_hdmx_data_from_base_to_composite_glyph",
    30: "dont_automatically_reorder_glyphs",
    46: "export_kern_table",
}

ttinfo_names = {
    0x33: "max_zones",
    0x34: "max_twilight_points",
    0x35: "max_storage",
    0x36: "max_function_defs",
    0x37: "max_instruction_defs",
    0x38: "max_stack_elements",
    0x39: "head_flags",  # tt_font_info_settings with head flags combined
    0x3A: "head_units_per_em",  # units_per_em, duplicate
    0x3B: "head_mac_style",
    0x3C: "head_lowest_rec_ppem",  # lowest_rec_ppem
    0x56: "head_creation",  # timestamp
    0x57: "_unknown_0x57",  # TODO: head_modification (not in API)?
    0x3D: "head_font_direction_hint",  # font_direction_hint
    0x3E: "os2_us_weight_class",  # weight_class, duplicate
    0x3F: "os2_us_width_class",  # width_class, duplicate
    0x40: "os2_fs_type",  # embedding
    0x41: "os2_y_subscript_x_size",  # subscript_x_size
    0x42: "os2_y_subscript_y_size",  # subscript_y_size
    0x43: "os2_y_subscript_x_offset",  # subscript_x_offset
    0x44: "os2_y_subscript_y_offset",  # subscript_y_offset
    0x45: "os2_y_superscript_x_size",  # superscript_x_size
    0x46: "os2_y_superscript_y_size",  # superscript_y_size
    0x47: "os2_y_superscript_x_offset",  # superscript_x_offset
    0x48: "os2_y_superscript_y_offset",  # superscript_y_offset
    0x49: "os2_y_strikeout_size",  # strikeout_size
    0x4A: "os2_y_strikeout_position",  # strikeout_position
    0x4B: "os2_s_family_class",  # ibm_classification + subclass
    0x4C: "OpenTypeOS2Panose",
    0x4D: "os2_s_typo_ascender",  # OpenTypeOS2TypoAscender
    0x4E: "os2_s_typo_descender",  # OpenTypeOS2TypoDescender
    0x4F: "os2_s_typo_line_gap",  # OpenTypeOS2TypoLineGap
    0x50: "os2_fs_selection",
    0x51: "os2_us_win_ascent",  # OpenTypeOS2WinAscent
    0x52: "os2_us_win_descent",  # OpenTypeOS2WinDescent
    0x5C: "Average Width",
    0x53: "Hdmx PPMs 1",
    0x58: "Hdmx PPMs 2",
    # os2_ul_code_page_range1, os2_ul_code_page_range2:
    0x54: "Codepages",
}

tt_settings = {
    0: "use_custom_tt_values",
    1: "create_vdmx",
    2: "add_null_cr_space",
}



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from vfbLib.compilers.header import VfbHeaderCompiler
from vfbLib.helpers import hexStr
from vfbLib.parsers.header import VfbHeaderParser

if TYPE_CHECKING:
    from io import BufferedReader


logger = logging.getLogger(__name__)


class VfbHeader:
    def __init__(self) -> None:
        # The original or compiled binary data
        self._data: bytes | None = None
        # The decompiled data
        self.decompiled: dict[str, Any] | None = None
        # Has the data been modified, i.e. it needs recompilation
        self.modified = False
        # The parser which can convert data to decompiled
        self.parser = VfbHeaderParser
        # The compiler which can convert the decompiled representation to bytes
        self.compiler = VfbHeaderCompiler

    def as_dict(self) -> dict[str, Any]:
        d: dict[str, Any] = {"decompiled": self.decompiled}
        if self.data is not None:
            d["data"] = hexStr(self.data)
        return d

    @property
    def data(self) -> bytes | None:
        return self._data

    @data.setter
    def data(self, value: bytes | None) -> None:
        if value != self._data:
            # New compiled data, we should remove the decompiled representation
            self.decompiled = None
        self._data = value

    def compile(self) -> None:
        """
        Compile the header. The result is stored in VfbHeader.data.
        """
        # XXX: Why do we need to always compile?
        # if not self.modified:
        #     return

        self._data = self.compiler().compile(self.decompiled)
        self.modified = False

    def read(self, stream: BufferedReader) -> None:
        self.decompiled = self.parser(stream).parse()



from unittest import TestCase

from fontTools.misc.textTools import deHexStr, hexStr

from vfbLib.compilers.text import NameRecordsCompiler, StringCompiler
from vfbLib.parsers.text import StringParser

decompiled_ibm_serif_text = [
    [0, 1, 0, 0, "Copyright 2020 IBM Corp. All rights reserved."],
    [0, 3, 1, 1033, "Copyright 2020 IBM Corp. All rights reserved."],
    [1, 1, 0, 0, "IBM Plex Serif Text"],
    [1, 3, 1, 1033, "IBM Plex Serif Text"],
    [2, 1, 0, 0, "Regular"],
    [2, 3, 1, 1033, "Regular"],
    [3, 1, 0, 0, "IBM;IBMPlexSerif-Text;2.006;2020"],
    [3, 3, 1, 1033, "IBM;IBMPlexSerif-Text;2.006;2020"],
    [4, 1, 0, 0, "IBM Plex Serif Text"],
    [4, 3, 1, 1033, "IBM Plex Serif Text"],
    [5, 1, 0, 0, "Version 2.006 2020"],
    [5, 3, 1, 1033, "Version 2.006 2020"],
    [6, 1, 0, 0, "IBMPlexSerif-Text"],
    [6, 3, 1, 1033, "IBMPlexSerif-Text"],
    [
        7,
        1,
        0,
        0,
        "IBM Plex® is a trademark of IBM Corp, registered in many jurisdictions worldwide.",
    ],
    [
        7,
        3,
        1,
        1033,
        "IBM Plex® is a trademark of IBM Corp, registered in many jurisdictions worldwide.",
    ],
    [8, 1, 0, 0, "Bold Monday"],
    [8, 3, 1, 1033, "Bold Monday"],
    [9, 1, 0, 0, "Mike Abbink, Paul van der Laan, Pieter van Rosmalen"],
    [9, 3, 1, 1033, "Mike Abbink, Paul van der Laan, Pieter van Rosmalen"],
    [11, 1, 0, 0, "http://www.boldmonday.com"],
    [11, 3, 1, 1033, "http://www.boldmonday.com"],
    [12, 1, 0, 0, "http://www.ibm.com"],
    [12, 3, 1, 1033, "http://www.ibm.com"],
    [
        13,
        1,
        0,
        0,
        "This Font Software is licensed under the SIL Open Font License, Version 1.1. This license is available with a FAQ at: http://scripts.sil.org/OFL",
    ],
    [
        13,
        3,
        1,
        1033,
        "This Font Software is licensed under the SIL Open Font License, Version 1.1. This license is available with a FAQ at: http://scripts.sil.org/OFL",
    ],
    [14, 1, 0, 0, "http://scripts.sil.org/OFL"],
    [14, 3, 1, 1033, "http://scripts.sil.org/OFL"],
    [16, 1, 0, 0, "IBM Plex Serif"],
    [16, 3, 1, 1033, "IBM Plex Serif"],
    [17, 1, 0, 0, "Text"],
    [17, 3, 1, 1033, "Text"],
    [19, 1, 0, 0, "How razorback-jumping frogs can level six piqued gymnasts!"],
    [19, 3, 1, 1033, "How razorback-jumping frogs can level six piqued gymnasts!"],
]
expected_ibm_serif_text = (
    "ad"  # 34
    # 0
    "8b8c8b8bb8cef703f704f70df706f4f2f3f708abbdbbbdbbabd4cdd8abcef703f706f704b9abccf700f700abf706f4f2f3f708f707abf706f0f707f0f706f70af0efb9"
    "8b8e8cfa9db8cef703f704f70df706f4f2f3f708abbdbbbdbbabd4cdd8abcef703f706f704b9abccf700f700abf706f4f2f3f708f707abf706f0f707f0f706f70af0efb9"
    # 1
    "8c8c8b8b9ed4cdd8abdbf700f0f70cabdef0f706f4f1abdff0f70cf708"
    "8c8e8cfa9d9ed4cdd8abdbf700f0f70cabdef0f706f4f1abdff0f70cf708"
    # 2
    "8d8c8b8b92ddf0f2f709f700ecf706"
    "8d8e8cfa9d92ddf0f2f709f700ecf706"
    # 3
    "8e8c8b8babd4cdd8c6d4cdd8dbf700f0f70cdef0f706f4f1b8dff0f70cf708c6bdb9bbbbc1c6bdbbbdbb"
    "8e8e8cfa9dabd4cdd8c6d4cdd8dbf700f0f70cdef0f706f4f1b8dff0f70cf708c6bdb9bbbbc1c6bdbbbdbb"
    # 4
    "8f8c8b8b9ed4cdd8abdbf700f0f70cabdef0f706f4f1abdff0f70cf708"
    "8f8e8cfa9d9ed4cdd8abdbf700f0f70cabdef0f706f4f1abdff0f70cf708"
    # 5
    "908c8b8b9de1f0f706f707f4f703f702abbdb9bbbbc1abbdbbbdbb"
    "908e8cfa9d9de1f0f706f707f4f703f702abbdb9bbbbc1abbdbbbdbb"
    # 6
    "918c8b8b9cd4cdd8dbf700f0f70cdef0f706f4f1b8dff0f70cf708"
    "918e8cfa9d9cd4cdd8dbf700f0f70cdef0f706f4f1b8dff0f70cf708"
    # 7
    "928c8b8bdcd4cdd8abdbf700f0f70cf73cabf4f707abecabf708f706eceff0f701ecf706f6abf703f1abd4cdd8abcef703f706f704b7abf706f0f2f4f707f708f0f706f0efabf4f702abf701ecf702f70dabf5f709f706f4f707eff4eef708f4f703f702f707abf70bf703f706f700eff70bf4eff0b9"
    "928e8cfa9ddcd4cdd8abdbf700f0f70cf742abf4f707abecabf708f706eceff0f701ecf706f6abf703f1abd4cdd8abcef703f706f704b7abf706f0f2f4f707f708f0f706f0efabf4f702abf701ecf702f70dabf5f709f706f4f707eff4eef708f4f703f702f707abf70bf703f706f700eff70bf4eff0b9"
    # 8
    "938c8b8b96cdf703f700efabd8f703f702efecf70d"
    "938e8cfa9d96cdf703f700efabd8f703f702efecf70d"
    # 9
    "948c8b8bbed8f4f6f0abccededf4f702f6b7abdbecf709f700abf70aecf702abeff0f706abd7ececf702b7abdbf4f0f708f0f706abf70aecf702abddf703f707f701ecf700f0f702"
    "948e8cfa9dbed8f4f6f0abccededf4f702f6b7abdbecf709f700abf70aecf702abeff0f706abd7ececf702b7abdbf4f0f708f0f706abf70aecf702abddf703f707f701ecf700f0f702"
    # 11
    "968c8b8ba4f3f708f708f704c5babaf70bf70bf70bb9edf703f700eff701f703f702efecf70db9eef703f701"
    "968e8cfa9da4f3f708f708f704c5babaf70bf70bf70bb9edf703f700eff701f703f702efecf70db9eef703f701"
    # 12
    "978c8b8b9df3f708f708f704c5babaf70bf70bf70bb9f4edf701b9eef703f701"
    "978e8cfa9d9df3f708f708f704c5babaf70bf70bf70bb9f4edf701b9eef703f701"
    # 13
    "988c8b8bf724dff3f4f707abd1f703f702f708abdef703f1f708f70becf706f0abf4f707abf700f4eef0f702f707f0efabf709f702eff0f706abf708f3f0abded4d7abdaf704f0f702abd1f703f702f708abd7f4eef0f702f707f0b7abe1f0f706f707f4f703f702abbcb9bcb9abdff3f4f707abf700f4eef0f702f707f0abf4f707abecf70aecf4f700ecedf700f0abf70bf4f708f3abecabd1ccdcabecf708c5abf3f708f708f704c5babaf707eef706f4f704f708f707b9f707f4f700b9f703f706f2badad1d7"
    "988e8cfa9df724dff3f4f707abd1f703f702f708abdef703f1f708f70becf706f0abf4f707abf700f4eef0f702f707f0efabf709f702eff0f706abf708f3f0abded4d7abdaf704f0f702abd1f703f702f708abd7f4eef0f702f707f0b7abe1f0f706f707f4f703f702abbcb9bcb9abdff3f4f707abf700f4eef0f702f707f0abf4f707abecf70aecf4f700ecedf700f0abf70bf4f708f3abecabd1ccdcabecf708c5abf3f708f708f704c5babaf707eef706f4f704f708f707b9f707f4f700b9f703f706f2badad1d7"
    # 14
    "998c8b8ba5f3f708f708f704c5babaf707eef706f4f704f708f707b9f707f4f700b9f703f706f2badad1d7"
    "998e8cfa9da5f3f708f708f704c5babaf707eef706f4f704f708f707b9f707f4f700b9f703f706f2badad1d7"
    # 16
    "9b8c8b8b99d4cdd8abdbf700f0f70cabdef0f706f4f1"
    "9b8e8cfa9d99d4cdd8abdbf700f0f70cabdef0f706f4f1"
    # 17
    "9c8c8b8b8fdff0f70cf708"
    "9c8e8cfa9d8fdff0f70cf708"
    # 19
    "9e8c8b8bc5d3f703f70babf706ecf70ef703f706edeceef6b8f5f709f701f704f4f702f2abf1f706f703f2f707abeeecf702abf700f0f70af0f700abf707f4f70cabf704f4f705f709f0efabf2f70df701f702ecf707f708f707ac"
    "9e8e8cfa9dc5d3f703f70babf706ecf70ef703f706edeceef6b8f5f709f701f704f4f702f2abf1f706f703f2f707abeeecf702abf700f0f70af0f700abf707f4f70cabf704f4f705f709f0efabf2f70df701f702ecf707f708f707ac"
)

raw_short = [
    [
        7,
        1,
        0,
        0,
        "IBM Plex® is a trademark of IBM Corp, registered in many jurisdictions worldwide.",
    ],
    [
        7,
        3,
        1,
        1033,
        "IBM Plex® is a trademark of IBM Corp, registered in many jurisdictions worldwide.",
    ],
]

expected_short = (
    "8d"  # 2
    # 7
    "928c8b8bdcd4cdd8abdbf700f0f70cf73cabf4f707abecabf708f706eceff0f701ecf706f6abf703f1abd4cdd8abcef703f706f704b7abf706f0f2f4f707f708f0f706f0efabf4f702abf701ecf702f70dabf5f709f706f4f707eff4eef708f4f703f702f707abf70bf703f706f700eff70bf4eff0b9"
    "928e8cfa9ddcd4cdd8abdbf700f0f70cf742abf4f707abecabf708f706eceff0f701ecf706f6abf703f1abd4cdd8abcef703f706f704b7abf706f0f2f4f707f708f0f706f0efabf4f702abf701ecf702f70dabf5f709f706f4f707eff4eef708f4f703f702f707abf70bf703f706f700eff70bf4eff0b9"
)


class NameRecordsCompilerTest(TestCase):
    def test_empty(self) -> None:
        result = NameRecordsCompiler().compile_hex([])
        assert result == "8b"

    def test_mac(self) -> None:
        result = NameRecordsCompiler().compile_hex([[0, 1, 0, 0, "Hällo"]])
        assert result == hexStr(deHexStr("8c 8b 8c 8b 8b 90 d3 f71e f700 f700 f703"))

    def test_unicode(self) -> None:
        result = NameRecordsCompiler().compile_hex([[0, 3, 1, 1033, "Hällo"]])
        assert result == hexStr(deHexStr("8c 8b 8e 8c fa9d 90 d3 f778 f700 f700 f703"))

    def test_ibm_serif_text(self) -> None:
        result = NameRecordsCompiler().compile_hex(decompiled_ibm_serif_text)
        assert result == expected_ibm_serif_text

    def test_mac_utf_encodings(self) -> None:
        result = NameRecordsCompiler().compile_hex(raw_short)
        assert result == expected_short


expected_regular = hexStr(deHexStr("52 65 67 75 6C 61 72"))
expected_space = hexStr(deHexStr("57 74 30 20 57 64 31 20"))


class StringCompilerTest(TestCase):
    def test_regular(self):
        result = StringCompiler().compile_hex("Regular")
        assert result == expected_regular

    def test_space(self):
        result = StringCompiler().compile_hex("Wt0 Wd1 ")  # Trailing space is compiled
        assert result == expected_space

    def test_space_roundtrip(self):
        dec = StringParser().parse_hex(expected_space)
        assert dec == "Wt0 Wd1 "  # Trailing space does not get stripped
        cde = StringCompiler().compile_hex(dec)
        assert cde == expected_space



import contextlib
import io
import sys
import unittest
from pathlib import Path
from unittest.mock import patch

# Import of the function to test
from vfbLib.cmdline import vfb2json, vfb2ufo


def data_path():
    return Path(__file__).parent / "Data"


def vfb_path(name: str) -> Path:
    return data_path() / name


class CmdlineTest(unittest.TestCase):
    def test_vfb2json(self):
        args = [
            "vfb2json",
            str(vfb_path("masters.vfb")),
            "-p",
            str(data_path() / "out"),
        ]
        with patch.object(sys, "argv", args):
            f = io.StringIO()
            with contextlib.redirect_stdout(f):
                vfb2json()
            # with open(font_path("css")) as expected:
            #     assert f.getvalue() == expected.read()

    def test_vfb2ufo(self):
        args = [
            "vfb3ufo",
            str(vfb_path("masters.vfb")),
            "-fo",
            "-p",
            str(data_path() / "out"),
        ]
        with patch.object(sys, "argv", args):
            f = io.StringIO()
            with contextlib.redirect_stdout(f):
                vfb2ufo()
            # with open(font_path("css")) as expected:
            #     assert f.getvalue() == expected.read()



from __future__ import annotations

import logging

from vfbLib import tt_settings, ttinfo_names
from vfbLib.helpers import binaryToIntList
from vfbLib.parsers.base import BaseParser
from vfbLib.typing import TTStemDict, TTStemsDict, TTZoneDict, TTZonesDict

logger = logging.getLogger(__name__)


def convert_int_to_flags_options(value: int) -> dict:
    flags = binaryToIntList(value & 0xFFFF)
    options = binaryToIntList(value >> 16)
    return {
        "flags": flags,
        "options": [tt_settings.get(i, i) for i in options],
    }


class TrueTypeInfoParser(BaseParser):
    """
    A parser that reads data as "TrueType Info" values.
    """

    def assert_unique(
        self, d: dict[str, int | list[int] | list[str]], key: str
    ) -> None:
        if key in d:
            logger.error(f"Duplicate key in TrueType Info: {key} in {d}")
            raise KeyError

    def _parse(self):
        info = {}

        while True:
            k = self.read_uint8()
            bk = ttinfo_names.get(k, f"bit{k}")  # dict key, as a bit number
            dk = ttinfo_names.get(k, str(k))  # dict key, human-readable
            hk = ttinfo_names.get(k, hex(k))  # dict key, hex

            if k == 0x32:
                return info

            elif k in (0x33, 0x34, 0x35, 0x36, 0x37, 0x38):
                self.assert_unique(info, dk)
                info[dk] = self.read_value()

            elif k == 0x39:
                # Options and head flags
                self.assert_unique(info, dk)
                info[dk] = convert_int_to_flags_options(self.read_value())

            elif k in (0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F):
                self.assert_unique(info, bk)
                info[bk] = self.read_value()

            elif k in (
                0x40,
                0x41,
                0x42,
                0x43,
                0x44,
                0x45,
                0x46,
                0x47,
                0x48,
                0x49,
                0x4A,
                0x4B,
            ):
                self.assert_unique(info, dk)
                info[dk] = self.read_value()

            elif k == 0x4C:  # PANOSE?
                self.assert_unique(info, dk)
                info[dk] = [self.read_uint8() for _ in range(10)]

            elif k in (0x4D, 0x4E, 0x4F, 0x50, 0x51, 0x52):
                self.assert_unique(info, dk)
                info[dk] = self.read_value()

            elif k == 0x53:
                self.assert_unique(info, dk)
                num_values = self.read_value()
                info[dk] = [self.read_uint8() for _ in range(num_values)]

            elif k == 0x54:
                # Codepages
                self.assert_unique(info, dk)
                range1 = self.read_value(signed=False)
                range2 = self.read_value(signed=False)
                info[dk] = {
                    "os2_ul_code_page_range1": range1,
                    "os2_ul_code_page_range2": range2,
                }

            elif k == 0x56:
                # Timestamp, unsigned
                self.assert_unique(info, dk)
                info[dk] = self.read_value(signed=False)

            elif k in (0x57, 0x5C):
                self.assert_unique(info, dk)
                info[dk] = self.read_value()

            elif k == 0x58:
                self.assert_unique(info, hk)
                num_values = self.read_value()
                info[hk] = [self.read_uint8() for _ in range(num_values)]

            else:
                logger.warning(f"Unknown key in TrueType info: {hex(k)}")


class TrueTypeStemsParser(BaseParser):
    def _parse(self):
        names = ("ttStemsV", "ttStemsH")
        result = TTStemsDict(ttStemsV=[], ttStemsH=[])
        for i in range(2):
            direction = []
            num_stems = self.read_value()
            for _ in range(num_stems):
                width = self.read_value()
                stem_name_length = self.read_uint8()
                stem_name = self.stream.read(stem_name_length).decode(self.encoding)
                ppm6 = self.read_value()

                direction.append(
                    TTStemDict(value=width, name=stem_name, round={"6": ppm6})
                )

            result[names[i]] = direction

        return result


class TrueTypeStemPpemsParser(BaseParser):
    def _parse(self):
        names = ("ttStemsV", "ttStemsH")
        result = TTStemsDict(ttStemsV=[], ttStemsH=[])
        for i in range(2):
            direction = []
            num_stems = self.read_value()
            d = {}
            for j in range(num_stems):
                for k in range(2, 6):
                    ppm = self.read_value()
                    d[str(k)] = ppm

                direction.append(TTStemDict(stem=j, round=d.copy()))

            result[names[i]] = direction

        return result


class TrueTypeStemPpems1Parser(BaseParser):
    # PPEM 1 for each stem is stored in a separate entry ...
    def _parse(self):
        names = ("ttStemsV", "ttStemsH")
        result = TTStemsDict(ttStemsV=[], ttStemsH=[])
        for i in range(2):
            direction = []
            num_stems = (self.ttStemsV_count, self.ttStemsH_count)[i]
            if num_stems is None:
                raise ValueError

            for j in range(num_stems):
                ppm = self.read_value()
                direction.append(TTStemDict(stem=j, round={"1": ppm}))

            result[names[i]] = direction

        return result


class TrueTypeStemPpems23Parser(BaseParser):
    # PPEM 2 and 3 for each stem are stored in a separate entry ... sometimes!?
    def _parse(self):
        names = ("ttStemsV", "ttStemsH")
        result = TTStemsDict(ttStemsV=[], ttStemsH=[])
        for i in range(2):
            direction = []
            num_stems = self.read_value()
            if num_stems is None:
                raise ValueError

            for j in range(num_stems):
                ppm2 = self.read_value()
                ppm3 = self.read_value()
                direction.append(TTStemDict(stem=j, round={"2": ppm2, "3": ppm3}))

            result[names[i]] = direction

        return result


class TrueTypeZoneDeltasParser(BaseParser):
    def _parse(self):
        num_deltas = self.read_value()
        result = {}
        for _ in range(num_deltas):
            # Index into Bottom + Top Zones
            index = self.read_value()
            ppm = self.read_value()
            shift = self.read_value()
            if index in result:
                result[index][ppm] = shift
            else:
                result[index] = {ppm: shift}

        return result


class TrueTypeZonesParser(BaseParser):
    def _parse(self):
        names = ("ttZonesT", "ttZonesB")
        result = TTZonesDict(ttZonesT=[], ttZonesB=[])
        for i in range(2):
            side = []
            num_zones = self.read_value()
            logger.debug(f"Zones: {num_zones}")
            for _ in range(num_zones):
                position = self.read_value()
                width = self.read_value()
                zone_name = self.read_str_with_len()
                side.append(TTZoneDict(position=position, value=width, name=zone_name))

            result[names[i]] = side

        return result


class VdmxParser(BaseParser):
    def _parse(self):
        result = []
        num_records = self.read_value()
        for _ in range(num_records):
            rec = {}
            rec["pelHeight"] = self.read_value()
            rec["max"] = self.read_value()
            rec["min"] = self.read_value()
            result.append(rec)

        return result



from io import BytesIO
from unittest import TestCase

import pytest
from fontTools.misc.textTools import deHexStr

from vfbLib.parsers.value import read_value


class ValueDecoderTest(TestCase):
    def expect(self, encoded, decoded):
        data = deHexStr(encoded)
        assert read_value(BytesIO(data)) == decoded

    def test_0x00(self):
        illegal = deHexStr("00")
        with pytest.raises(EOFError):
            read_value(BytesIO(illegal))

    def test_0x19(self):
        illegal = deHexStr("19")
        with pytest.raises(ValueError):
            read_value(BytesIO(illegal))

    def test_0x20(self):
        self.expect("20", -107)

    def test_0x8b(self):
        self.expect("8b", 0)

    def test_0xf6(self):
        self.expect("f6", 107)

    def test_0xf700(self):
        self.expect("f700", 108)

    def test_0xf701(self):
        self.expect("f701", 109)

    def test_0xf7ff(self):
        self.expect("f7ff", 363)

    def test_0xf800(self):
        self.expect("f800", 364)

    def test_0xf801(self):
        self.expect("f801", 365)

    def test_0xfa00(self):
        self.expect("fa00", 876)

    def test_0xfaff(self):
        self.expect("faff", 1131)

    def test_0xfb00(self):
        self.expect("fb00", -108)

    def test_0xfb01(self):
        self.expect("fb01", -109)

    def test_0xfe00(self):
        self.expect("fe00", -876)

    def test_0xfeff(self):
        self.expect("feff", -1131)

    def test_0xff00000000(self):
        self.expect("ff00000000", 0)

    def test_0xff00001000(self):
        self.expect("ff00001000", 4096)

    def test_0xffffffffff(self):
        self.expect("ffffffffff", -1)

    def test_0xffffffefff(self):
        self.expect("ffffffefff", -4097)



from __future__ import annotations

# try:
#     # NotRequired is only available in Python 3.11+
#     from typing import NotRequired, TypedDict
# except ImportError:
from typing import Any, Literal

from typing_extensions import NotRequired, TypedDict

Point = tuple[int, int]


class AnchorDict(TypedDict):
    name: NotRequired[str]
    x: int
    x1: NotRequired[int]
    y: int
    y1: NotRequired[int]


KerningClassFlagDict = dict[str, tuple[int, int]]

MetricsClassFlagDict = dict[str, tuple[int, int, int]]


class BBoxDict(TypedDict):
    xMin: int
    yMin: int
    xMax: int
    yMax: int


class Component(TypedDict):
    gid: int
    offsetX: list[int]
    offsetY: list[int]
    scaleX: list[float]
    scaleY: list[float]


class CustomCmap(TypedDict):
    language_id: int
    platform_id: int
    encoding_id: int
    format: int
    option: int
    records: list[int]


EntryDecompiled = dict[str, Any] | int | list[Any] | str | tuple[int, str] | None


class FeatureDict(TypedDict):
    tag: str
    code: list[str]


class FeaturesDict(TypedDict):
    prefix: list[str]
    features: list[FeatureDict]


GaspList = list[dict[str, int]]


class GdefDict(TypedDict):
    anchors: NotRequired[list[AnchorDict]]
    carets: NotRequired[list[tuple[int, int]]]
    glyph_class: NotRequired[str]
    unknown: NotRequired[list[int]]


class GuideDict(TypedDict):
    angle: float | int
    pos: int


GuideList = list[GuideDict]


class MMGuidesDict(TypedDict):
    h: list[GuideList]
    v: list[GuideList]


class GuidePropertyDict(TypedDict):
    color: NotRequired[str]
    index: int
    name: NotRequired[str]


class GuidePropertiesDict(TypedDict):
    h: list[GuidePropertyDict]
    v: list[GuidePropertyDict]


class HintDict(TypedDict):
    pos: int
    width: int


HintTuple = tuple[str, int, int]


class Instruction(TypedDict):
    cmd: str
    params: dict[str, int]


class LinkDict(TypedDict):
    x: list[tuple[int, int]]
    y: list[tuple[int, int]]


class MMAnchorDict(TypedDict):
    x: list[int]
    y: list[int]


class MMHintsDict(TypedDict):
    h: list[list[HintDict]]
    v: list[list[HintDict]]
    hintmasks: NotRequired[list[tuple[str, int]]]


class MMNode(TypedDict):
    flags: int
    points: list[list[Point]]
    type: Literal["move", "line", "curve", "qcurve"]


class PSInfoDict(TypedDict):
    font_matrix: tuple[float, ...]
    force_bold: int
    blue_values: list[int]
    other_blues: list[int]
    family_blues: list[int]
    family_other_blues: list[int]
    blue_scale: float
    blue_shift: int
    blue_fuzz: int
    std_hw: int
    std_vw: int
    stem_snap_h: list[int]
    stem_snap_v: list[int]
    bounding_box: BBoxDict
    adv_width_min: int
    adv_width_max: int
    adv_width_avg: int
    ascender: int
    descender: int
    x_height: int
    cap_height: int


class TTCommandDict(TypedDict):
    name: str
    params: list[str]


class TTStemDict(TypedDict):
    name: NotRequired[str]
    round: dict[str, int]
    stem: NotRequired[int]
    value: NotRequired[int]


class TTStemsDict(TypedDict):
    ttStemsV: list[TTStemDict]
    ttStemsH: list[TTStemDict]


class TTZoneDict(TypedDict):
    deltas: NotRequired[dict[int, int]]
    name: str
    position: int
    top: NotRequired[bool]
    value: int


class TTZonesDict(TypedDict):
    ttZonesT: list[TTZoneDict]
    ttZonesB: list[TTZoneDict]


class GlyphData(TypedDict):
    components: NotRequired[list[Component]]
    # constants: NotRequired[Tuple[Any, ...]]
    guides: NotRequired[MMGuidesDict]
    hints: NotRequired[MMHintsDict]
    imported: NotRequired[Any]  # FIXME
    kerning: NotRequired[dict[int, list[int]]]
    metrics: NotRequired[list[Point]]
    name: NotRequired[str]
    nodes: NotRequired[list[MMNode]]
    num_masters: NotRequired[int]
    # num_node_values: NotRequired[int]
    tth: NotRequired[list[Instruction]]


class MaskData(GlyphData):
    num: int
    reserved0: NotRequired[int]


class VfbDict(TypedDict):
    header: dict[str, Any]
    entries: list[dict[str, Any]]



import contextlib
import io
import sys
import unittest
from pathlib import Path
from unittest.mock import patch

# Import of the function to test
from vfbLib.cu2qu import vfbcu2qu


def data_path():
    return Path(__file__).parent / "Data"


def vfb_path(name: str) -> Path:
    return data_path() / name


class Cu2quTest(unittest.TestCase):
    def test_vfbcu2qu(self):
        args = [
            "vfbcu2qu",
            str(vfb_path("cu2qu.vfb")),
            "-fo",
            "-p",
            str(data_path() / "out"),
        ]
        with patch.object(sys, "argv", args):
            f = io.StringIO()
            with contextlib.redirect_stdout(f):
                vfbcu2qu()
            # with open(font_path("css")) as expected:
            #     assert f.getvalue() == expected.read()

    def test_vfbcu2qu_2_masters(self):
        args = [
            "vfbcu2qu",
            str(vfb_path("cu2qu_2m.vfb")),
            "-fo",
            "-p",
            str(data_path() / "out"),
        ]
        with patch.object(sys, "argv", args):
            f = io.StringIO()
            with contextlib.redirect_stdout(f):
                vfbcu2qu()
            # with open(font_path("css")) as expected:
            #     assert f.getvalue() == expected.read()



from unittest import TestCase

from vfbLib.parsers.base import (
    EncodedValueListParser,
    OpenTypeKerningClassFlagsParser,
    OpenTypeMetricsClassFlagsParser,
)


class EncodedValueListParserTest(TestCase):
    def test_links_1(self):
        data = "8c 8d 89 8b"
        expected = [1, 2, -2, 0]
        result = EncodedValueListParser().parse_hex(data)
        assert result == expected


class OpenTypeClassFlagsParserTest(TestCase):
    def test_2_masters_kerning(self):
        # fmt: off
        data = (
            "8e"                # 3 entries follow

            # entry 1
            "92"                # 7 = string length
            "5f615f4c454654"    # _a_LEFT
            "fa94"              # 1024
            "8b"                # 0

            # entry 2
            "93"                # 8 = string length
            "5f615f5249474854"  # _a_RIGHT
            "ff00000800"        # 2048
            "8b"                # 0

            # entry 3
            "8d"                # 2 = string length
            "5f74"              # _t
            "ff00000c00"        # 3072
            "8b"                # 0
        )
        # fmt: on
        expected = {"_a_LEFT": (1024, 0), "_a_RIGHT": (2048, 0), "_t": (3072, 0)}
        result = OpenTypeKerningClassFlagsParser().parse_hex(data)
        assert result == expected

    def test_2_masters_metrics(self):
        # fmt: off
        data = (
            "8c"            # 1 entry follows

            # entry 1
            "91"            # 6 = string length
            "2e6d74727835"  # .mtrx5
            "8b"            # 0
            "ff00001401"    # 5121
            "8b"            # 0
        )
        # fmt: on
        expected = {".mtrx5": (0, 5121, 0)}

        result = OpenTypeMetricsClassFlagsParser().parse_hex(data)
        assert result == expected



from __future__ import annotations

import logging
from io import BytesIO
from typing import Any

from vfbLib.compilers.base import StreamWriter

logger = logging.getLogger(__name__)


class VfbHeaderCompiler(StreamWriter):
    encoding = "cp1252"

    def compile(self, data: Any) -> bytes:
        self.stream = BytesIO()
        self._compile(data)
        return self.stream.getvalue()

    def _compile(self, data: Any) -> None:
        self.write_uint8(data["header0"])
        self.write_str(data["filetype"])
        self.write_uint16(data["header1"])

        # Compile chunk1 separately so we can store its size
        sw = StreamWriter()
        sw.stream = BytesIO()
        for value in data["chunk1"]:
            sw.write_uint8(value)
        chunk1 = sw.stream.getvalue()
        self.write_uint16(len(chunk1))
        self.write_bytes(chunk1)
        if data["chunk1"][-2:] == [10, 0]:
            # FL4+ additions over FL3
            # Compile the app_info chunk separately so we can store its size
            sw = StreamWriter()
            sw.stream = BytesIO()
            for k, v in data["creator"].items():
                sw.write_uint8(int(k))
                if int(k) == 2:
                    # app version
                    sw.write_uint8(0xFF)
                    for version in v:
                        sw.write_uint8(version)
                else:
                    sw.write_value(v)
            sw.write_uint8(0)  # stop marker of the key-value dict
            app_info = sw.stream.getvalue()
            self.write_uint16(len(app_info))
            self.write_bytes(app_info)
            # In the older header, this is already at the end of chunk1:
            self.write_uint8(6)
            self.write_uint8(1)
        self.write_uint16(0)



from unittest import TestCase

from vfbLib.tth import (
    extract_tt_stems,
    extract_tt_zones,
)


class TTHExportTest(TestCase):
    def test_extract_tt_stems(self):
        data = {
            "ttStemsV": [
                {"value": 77, "name": "currency_stroke", "round": {"6": 72}},
                {"value": 84, "name": "currency_white", "round": {"6": 66}},
            ],
            "ttStemsH": [
                {"value": 109, "name": "X: 109", "round": {"6": 51}},
                {"value": 91, "name": "X: 91", "round": {"6": 61}},
            ],
        }
        target = {}
        extract_tt_stems(data, target)
        assert target["stems"] == {
            "ttStemsV": [
                {
                    "width": 77,
                    "name": "currency_stroke",
                    "round": {"6": 72},
                    "horizontal": True,
                },
                {
                    "width": 84,
                    "name": "currency_white",
                    "round": {"6": 66},
                    "horizontal": True,
                },
            ],
            "ttStemsH": [
                {
                    "width": 109,
                    "name": "X: 109",
                    "round": {"6": 51},
                    "horizontal": False,
                },
                {
                    "width": 91,
                    "name": "X: 91",
                    "round": {"6": 61},
                    "horizontal": False,
                },
            ],
        }

    def test_extract_tt_stems_duplicate_names(self):
        data = {
            "ttStemsV": [
                {"value": 77, "name": "currency", "round": {"6": 72}},
                {"value": 84, "name": "currency", "round": {"6": 66}},
            ],
            "ttStemsH": [
                {"value": 72, "name": "currency", "round": {"6": 74}},
            ],
        }
        target = {}
        extract_tt_stems(data, target)
        assert target["stems"]["ttStemsV"] == [
            {
                "width": 77,
                "name": "currency",
                "round": {"6": 72},
                "horizontal": True,
            },
            {
                "width": 84,
                "name": "currency#1",
                "round": {"6": 66},
                "horizontal": True,
            },
        ]
        assert target["stems"]["ttStemsH"] == [
            {
                "width": 72,
                "name": "currency#2",
                "round": {"6": 74},
                "horizontal": False,
            },
        ]

    def test_extract_tt_zones(self):
        data = {
            "ttZonesT": [
                {"position": 520, "value": 12, "name": "xheight"},
            ],
            "ttZonesB": [
                {"position": 0, "value": 12, "name": "baseline"},
            ],
        }
        target = {}
        zone_names = {"ttZonesT": {}, "ttZonesB": {}}
        extract_tt_zones(data, target, zone_names)
        assert zone_names["ttZonesT"] == {0: "xheight"}
        assert zone_names["ttZonesB"] == {0: "baseline"}

    def test_extract_tt_zones_duplicate_names(self):
        data = {
            "ttZonesT": [
                {"position": 520, "value": 12, "name": "xheight"},
                {"position": 530, "value": 12, "name": "xheight"},
            ],
            "ttZonesB": [
                {"position": 0, "value": 12, "name": "xheight"},
            ],
        }
        target = {}
        zone_names = {"ttZonesT": {}, "ttZonesB": {}}
        extract_tt_zones(data, target, zone_names)
        assert zone_names["ttZonesT"] == {0: "xheight", 1: "xheight#1"}
        assert zone_names["ttZonesB"] == {0: "xheight#2"}



from __future__ import annotations

from vfbLib.parsers.base import BaseParser
from vfbLib.typing import CustomCmap


class CustomCmapParser(BaseParser):
    """
    A parser that reads the custom CMAP settings.
    """

    def _parse(self) -> list[CustomCmap]:
        count = self.read_value()  # number of cmap records, not values
        values = []
        for _ in range(count):
            cmap = CustomCmap(
                language_id=self.read_value(),
                platform_id=self.read_value(),
                encoding_id=self.read_value(),
                format=self.read_value(),
                option=self.read_value(),
                records=[],
            )
            num_encodings = self.read_value()
            if num_encodings > 0:
                cmap["records"] = [self.read_uint8() for _ in range(num_encodings)]
            values.append(cmap)
        return values



from __future__ import annotations

from vfbLib.compilers.base import (
    EncodedValueListCompiler,
    EncodedValueListWithCountCompiler,
    GlyphEncodingCompiler,
    HexStringCompiler,
    MappingModeCompiler,
    OpenTypeKerningClassFlagsCompiler,
    OpenTypeMetricsClassFlagsCompiler,
)
from vfbLib.compilers.binary import BinaryTableCompiler
from vfbLib.compilers.cmap import CustomCmapCompiler
from vfbLib.compilers.glyph import (
    GlyphAnchorsCompiler,
    GlyphAnchorsSuppCompiler,
    GlyphCompiler,
    GlyphGDEFCompiler,
    GlyphOriginCompiler,
    GlyphUnicodesCompiler,
    GlyphUnicodesSuppCompiler,
    LinksCompiler,
    MaskCompiler,
    MaskMetricsCompiler,
    MaskMetricsMMCompiler,
)
from vfbLib.compilers.guides import GuidePropertiesCompiler, GuidesCompiler
from vfbLib.compilers.mm import (
    AnisotropicInterpolationsCompiler,
    AxisMappingsCompiler,
    AxisMappingsCountCompiler,
    MasterLocationCompiler,
    PrimaryInstancesCompiler,
)
from vfbLib.compilers.numeric import (
    DoubleCompiler,
    DoubleListCompiler,
    Int16Compiler,
    IntListCompiler,
    PanoseCompiler,
    SignedInt16Compiler,
    SignedInt32Compiler,
    UnicodeRangesCompiler,
)
from vfbLib.compilers.options import (
    ExportOptionsCompiler,
    OpenTypeExportOptionsCompiler,
)
from vfbLib.compilers.pclt import PcltCompiler
from vfbLib.compilers.ps import (
    PostScriptGlobalHintingOptionsCompiler,
    PostScriptGlyphHintingOptionsCompiler,
    PostScriptInfoCompiler,
)
from vfbLib.compilers.text import (
    NameRecordsCompiler,
    OpenTypeStringCompiler,
    StringCompiler,
    VendorIdCompiler,
)
from vfbLib.compilers.truetype import (
    GaspCompiler,
    TrueTypeInfoCompiler,
    TrueTypeStemPpems1Compiler,
    TrueTypeStemPpems23Compiler,
    TrueTypeStemPpemsCompiler,
    TrueTypeStemsCompiler,
    TrueTypeZoneDeltasCompiler,
    TrueTypeZonesCompiler,
    VdmxCompiler,
)
from vfbLib.parsers.base import (  # EncodedValueParser,; EncodedKeyValuesParser,
    BaseParser,
    EncodedValueListParser,
    EncodedValueListWithCountParser,
    GaspParser,  # TODO: Should move to parsers.truetype
    GlyphEncodingParser,
    MappingModeParser,
    OpenTypeKerningClassFlagsParser,
    OpenTypeMetricsClassFlagsParser,
)
from vfbLib.parsers.binary import BinaryTableParser
from vfbLib.parsers.bitmap import BackgroundBitmapParser, GlyphBitmapParser
from vfbLib.parsers.cmap import CustomCmapParser
from vfbLib.parsers.fl3 import FL3Type1410Parser
from vfbLib.parsers.glyph import (
    GlobalMaskParser,
    GlyphAnchorsParser,
    GlyphAnchorsSuppParser,
    GlyphGDEFParser,
    GlyphOriginParser,
    GlyphParser,
    GlyphSketchParser,
    GlyphUnicodeParser,
    GlyphUnicodeSuppParser,
    LinkParser,
    MaskMetricsMMParser,
    MaskMetricsParser,
    MaskParser,
)
from vfbLib.parsers.guides import GlobalGuidesParser, GuidePropertiesParser
from vfbLib.parsers.mm import (
    AnisotropicInterpolationsParser,
    AxisMappingsCountParser,
    AxisMappingsParser,
    MasterLocationParser,
    PrimaryInstancesParser,
)
from vfbLib.parsers.numeric import (
    DoubleListParser,
    DoubleParser,
    Int16Parser,
    IntListParser,
    PanoseParser,
    SignedInt16Parser,
    SignedInt32Parser,
    UnicodeRangesParser,
)
from vfbLib.parsers.options import ExportOptionsParser, OpenTypeExportOptionsParser
from vfbLib.parsers.pclt import PcltParser
from vfbLib.parsers.ps import (
    PostScriptGlobalHintingOptionsParser,
    PostScriptGlyphHintingOptionsParser,
    PostScriptInfoParser,
)
from vfbLib.parsers.text import NameRecordsParser, OpenTypeStringParser, StringParser
from vfbLib.parsers.truetype import (
    TrueTypeInfoParser,
    TrueTypeStemPpems1Parser,
    TrueTypeStemPpems23Parser,
    TrueTypeStemPpemsParser,
    TrueTypeStemsParser,
    TrueTypeZoneDeltasParser,
    TrueTypeZonesParser,
    VdmxParser,
)

# fmt: off
parser_classes = {
    # Sorted by appearance in the VFB
    1501: ("Encoding Default", GlyphEncodingParser, GlyphEncodingCompiler),
    1500: ("Encoding", GlyphEncodingParser, GlyphEncodingCompiler),
    1502: ("1502", Int16Parser, Int16Compiler),
    518: ("518", StringParser, StringCompiler),
    257: ("257", StringParser, StringCompiler),
    1026: ("font_name", StringParser, StringCompiler),
    1503: ("Master Count", Int16Parser, Int16Compiler),
    1517: ("weight_vector", DoubleListParser, DoubleListCompiler),
    1044: ("unique_id", SignedInt32Parser, SignedInt32Compiler),
    1046: ("version", StringParser, StringCompiler),
    1038: ("notice", StringParser, StringCompiler),
    1025: ("full_name", StringParser, StringCompiler),
    1027: ("family_name", StringParser, StringCompiler),
    1024: ("pref_family_name", StringParser, StringCompiler),
    1056: ("menu_name", StringParser, StringCompiler),
    1092: ("apple_name", StringParser, StringCompiler),
    1028: ("weight", StringParser, StringCompiler),
    1065: ("width", StringParser, StringCompiler),

    # Is license/url not in Python API?
    1069: ("License", StringParser, StringCompiler),
    1070: ("License URL", StringParser, StringCompiler),

    1037: ("copyright", StringParser, StringCompiler),
    1061: ("trademark", StringParser, StringCompiler),
    1062: ("designer", StringParser, StringCompiler),
    1063: ("designer_url", StringParser, StringCompiler),
    1064: ("vendor_url", StringParser, StringCompiler),
    1039: ("source", StringParser, StringCompiler),  # manufacturer, "created by"
    1034: ("is_fixed_pitch", Int16Parser, Int16Compiler),
    1048: ("weight_code", SignedInt16Parser, SignedInt16Compiler),
    1029: ("italic_angle", DoubleParser, DoubleCompiler),
    1047: ("slant_angle", DoubleParser, DoubleCompiler),
    1030: ("underline_position", SignedInt16Parser, SignedInt16Compiler),
    1031: ("underline_thickness", Int16Parser, Int16Compiler),
    1054: ("ms_charset", Int16Parser, Int16Compiler),
    1118: ("panose", PanoseParser, PanoseCompiler),
    1128: ("tt_version", StringParser, StringCompiler),
    1129: ("tt_u_id", StringParser, StringCompiler),
    1127: ("style_name", StringParser, StringCompiler),
    1137: ("pref_style_name", StringParser, StringCompiler),
    1139: ("mac_compatible", StringParser, StringCompiler),
    1140: ("1140", BaseParser, HexStringCompiler),
    1121: ("vendor", StringParser, VendorIdCompiler),
    1133: ("xuid", IntListParser, IntListCompiler),
    1134: ("xuid_num", Int16Parser, Int16Compiler),
    1132: ("year", Int16Parser, Int16Compiler),
    1130: ("version_major", Int16Parser, Int16Compiler),
    1131: ("version_minor", Int16Parser, Int16Compiler),
    1135: ("upm", Int16Parser, Int16Compiler),
    1090: ("fond_id", Int16Parser, Int16Compiler),
    1093: ("PostScript Hinting Options", PostScriptGlobalHintingOptionsParser, PostScriptGlobalHintingOptionsCompiler),  # noqa: E501
    1068: ("1068", EncodedValueListWithCountParser, EncodedValueListWithCountCompiler),
    1530: ("blue_values_num", Int16Parser, Int16Compiler),
    1531: ("other_blues_num", Int16Parser, Int16Compiler),
    1532: ("family_blues_num", Int16Parser, Int16Compiler),
    1533: ("family_other_blues_num", Int16Parser, Int16Compiler),
    1534: ("stem_snap_h_num", Int16Parser, Int16Compiler),
    1535: ("stem_snap_v_num", Int16Parser, Int16Compiler),
    1267: ("font_style", Int16Parser, Int16Compiler),  # OS/2.fsSelection
    1057: ("pcl_id", Int16Parser, Int16Compiler),
    1058: ("vp_id", Int16Parser, Int16Compiler),
    1060: ("ms_id", Int16Parser, Int16Compiler),
    1059: ("pcl_chars_set", StringParser, StringCompiler),

    # Goes to font.ttinfo:
    1261: ("cvt", BaseParser, HexStringCompiler),  # Binary cvt Table
    1262: ("prep", BaseParser, HexStringCompiler),  # Binary prep Table
    1263: ("fpgm", BaseParser, HexStringCompiler),  # Binary fpgm Table
    1265: ("gasp", GaspParser, GaspCompiler),
    1264: ("ttinfo", TrueTypeInfoParser, TrueTypeInfoCompiler),
    # Goes to font.ttinfo:
    1271: ("vdmx", VdmxParser, VdmxCompiler),
    1270: ("hhea_line_gap", Int16Parser, Int16Compiler),
    1278: ("hhea_ascender", SignedInt16Parser, SignedInt16Compiler),
    1279: ("hhea_descender", SignedInt16Parser, SignedInt16Compiler),
    # hstem_data and vstem_data, goes to font.ttinfo:
    1266: ("TrueType Stem PPEMs 2 And 3", TrueTypeStemPpems23Parser, TrueTypeStemPpems23Compiler),  # noqa: E501
    1268: ("TrueType Stem PPEMs", TrueTypeStemPpemsParser, TrueTypeStemPpemsCompiler),
    # Probably in font.ttinfo, but not accessible through API:
    1269: ("TrueType Stems", TrueTypeStemsParser, TrueTypeStemsCompiler),
    1524: ("TrueType Stem PPEMs 1", TrueTypeStemPpems1Parser, TrueTypeStemPpems1Compiler),  # noqa: E501
    # Probably in font.ttinfo, but not accessible through API:
    1255: ("TrueType Zones", TrueTypeZonesParser, TrueTypeZonesCompiler),

    # Goes to font:
    2021: ("unicoderanges", UnicodeRangesParser, UnicodeRangesCompiler),

    # Probably in font.ttinfo, but not accessible through API:
    1272: ("stemsnaplimit", Int16Parser, Int16Compiler),  # Pixel Snap
    1274: ("zoneppm", Int16Parser, Int16Compiler),  # Zone Stop PPEM
    1275: ("codeppm", Int16Parser, Int16Compiler),  # Code Stop PPEM
    1604: ("1604", Int16Parser, Int16Compiler),  # Binary import? e.g. 255
    2032: ("2032", Int16Parser, Int16Compiler),  # Binary import? e.g. 300
    1273: ("TrueType Zone Deltas", TrueTypeZoneDeltasParser, TrueTypeZoneDeltasCompiler),  # noqa: E501

    # Goes to font again:
    1138: ("fontnames", NameRecordsParser, NameRecordsCompiler),
    1141: ("Custom CMAPs", CustomCmapParser, CustomCmapCompiler),
    1136: ("PCLT Table", PcltParser, PcltCompiler),
    2022: ("Export PCLT Table", Int16Parser, Int16Compiler),
    2025: ("note", StringParser, StringCompiler),
    2030: ("2030", BaseParser, HexStringCompiler),
    2016: ("customdata", StringParser, StringCompiler),
    2024: ("OpenType Metrics Class Flags", OpenTypeMetricsClassFlagsParser, OpenTypeMetricsClassFlagsCompiler),  # noqa: E501
    2026: ("OpenType Kerning Class Flags", OpenTypeKerningClassFlagsParser, OpenTypeKerningClassFlagsCompiler),  # noqa: E501

    # Repeat for each binary table:
    # truetypetables: TrueTypeTable
    2014: ("TrueTypeTable", BinaryTableParser, BinaryTableCompiler),

    1276: ("features", OpenTypeStringParser, OpenTypeStringCompiler),

    # Repeat for each OpenType class:
    1277: ("OpenType Class", StringParser, StringCompiler),  # Font.classes

    513: ("513", BaseParser, HexStringCompiler),
    271: ("271", BaseParser, HexStringCompiler),
    1513: ("Axis Count", Int16Parser, Int16Compiler),
    1514: ("Axis Name", StringParser, StringCompiler),
    1523: ("Anisotropic Interpolation Mappings", AnisotropicInterpolationsParser, AnisotropicInterpolationsCompiler),  # noqa: E501
    1515: ("Axis Mappings Count", AxisMappingsCountParser, AxisMappingsCountCompiler),
    1516: ("Axis Mappings", AxisMappingsParser, AxisMappingsCompiler),

    # Repeat the next two for each master:
    1504: ("Master Name", StringParser, StringCompiler),
    1505: ("Master Location", MasterLocationParser, MasterLocationCompiler),

    1247: ("Primary Instance Locations", DoubleListParser, DoubleListCompiler),
    1254: ("Primary Instances", PrimaryInstancesParser, PrimaryInstancesCompiler),

    # Repeat PostScript Info for each master:
    1536: ("PostScript Info", PostScriptInfoParser, PostScriptInfoCompiler),

    527: ("527", BaseParser, HexStringCompiler),
    1294: ("Global Guides", GlobalGuidesParser, GuidesCompiler),
    1296: ("Global Guide Properties", GuidePropertiesParser, GuidePropertiesCompiler),
    1295: ("Global Mask", GlobalMaskParser, None),
    1066: ("default_character", StringParser, StringCompiler),

    # Begin: Repeat for each glyph
    2001: ("Glyph", GlyphParser, GlyphCompiler),
    # Glyph.hlinks and Glyph.vlinks:
    2008: ("Links", LinkParser, LinksCompiler),
    2007: ("image", BackgroundBitmapParser, None),  # Background Bitmap
    2013: ("Glyph Bitmaps", GlyphBitmapParser, None),
    2023: ("2023", EncodedValueListParser, EncodedValueListCompiler),  # 1 encoded value per master  # noqa: E501
    2019: ("Glyph Sketch", GlyphSketchParser, None),
    2010: ("Glyph Hinting Options", PostScriptGlyphHintingOptionsParser, PostScriptGlyphHintingOptionsCompiler),  # noqa: E501
    2009: ("mask", MaskParser, MaskCompiler),
    2011: ("mask.metrics", MaskMetricsParser, MaskMetricsCompiler),  # Single master mask metrics  # noqa: E501
    2028: ("mask.metrics_mm", MaskMetricsMMParser, MaskMetricsMMCompiler),  # Mask metrics master 2 to 16  # noqa: E501
    2027: ("Glyph Origin", GlyphOriginParser, GlyphOriginCompiler),
    1250: ("unicodes", GlyphUnicodeParser, GlyphUnicodesCompiler),  # Glyph Unicode
    2034: ("2034", StringParser, StringCompiler),
    1253: ("Glyph Unicode Non-BMP", GlyphUnicodeSuppParser, GlyphUnicodesSuppCompiler),
    2012: ("mark", Int16Parser, Int16Compiler),  # Mark Color
    2015: ("glyph.customdata", StringParser, StringCompiler),
    2017: ("glyph.note", StringParser, StringCompiler),
    2018: ("Glyph GDEF Data", GlyphGDEFParser, GlyphGDEFCompiler),
    2020: ("Glyph Anchors Supplemental", GlyphAnchorsSuppParser, GlyphAnchorsSuppCompiler),  # noqa: E501
    2029: ("Glyph Anchors MM", GlyphAnchorsParser, GlyphAnchorsCompiler),  # MM-compatible  # noqa: E501
    2031: ("Glyph Guide Properties", GuidePropertiesParser, GuidePropertiesCompiler),
    # End: Repeat for each glyph

    1743: ("OpenType Export Options", OpenTypeExportOptionsParser, OpenTypeExportOptionsCompiler),  # noqa: E501
    1744: ("Export Options", ExportOptionsParser, ExportOptionsCompiler),
    1742: ("Mapping Mode", MappingModeParser, MappingModeCompiler),

    # Not seen in FontNames.vfb:
    1410: ("1410", FL3Type1410Parser, None),

    # File end
    5: ("EOF", None, None),
}
# fmt: on

# Make sure the human-readable keys are unique
all_classes = [key for key, _, _ in parser_classes.values()]
assert len(set(all_classes)) == len(all_classes), (
    f"Duplicate keys in classes: {sorted(all_classes)}"
)

entry_ids = {v[0]: k for k, v in parser_classes.items()}

# Those entries are ignored in minimal mode:
ignore_minimal = [
    "Global Guides",
    "Global Mask",
    "Glyph Bitmaps",
    "Glyph Guide Properties",
    "glyph.note",
    "image",
    "mark",
    "mask",
    "note",
]
